import logging
import os
import subprocess
import sys
from threading import Event, Thread
from typing import Sequence

from hhd.i18n import _
from hhd.plugins import Context, HHDPlugin, HHDSettings, load_relative_yaml
from hhd.plugins.conf import Config
from hhd.utils import GIT_ADJ, GIT_HHD, HHD_DEV_DIR

from .logs import get_log

logger = logging.getLogger(__name__)

FPASTE_SERVICE = os.environ.get("HHD_FPASTE", "fpaste")
BUGREPORTS_ENABLED = os.environ.get("HHD_BUGREPORT", "0") == "1"
USES_BETA = os.environ.get("HHD_SWITCH_ROOT", "0") == "1"


def prepare_hhd_dev(ev):
    try:
        os.makedirs(HHD_DEV_DIR, exist_ok=True)
        subprocess.run(
            ["python3", "-m", "venv", "--system-site-packages", HHD_DEV_DIR], check=True
        )
        subprocess.run(
            [
                f"{HHD_DEV_DIR}/bin/pip",
                "install",
                "--upgrade",
                "--cache-dir",
                "/tmp/__hhd_update_cache",
                GIT_HHD,
                GIT_ADJ,
            ],
            check=True,
        )
    except Exception as e:
        ev.set()
        # Show full stacktrace
        raise e


def upload_log(boot: str, out: dict):
    match boot:
        case "current":
            bootnum = 0
        case "previous":
            bootnum = -1
        case "m2":
            bootnum = -2
        case "m3":
            bootnum = -3
        case _:
            bootnum = 0

    logs = get_log(bootnum)
    try:
        res = subprocess.run(
            [FPASTE_SERVICE, "-x", str(72 * 60)],
            input=logs,
            check=True,
            capture_output=True,
            text=True,
        )
        if res.returncode != 0:
            raise Exception(f"fpaste failed with code {res.returncode}")
        
        out["url"] = res.stdout.strip()
        out["error"] = None
    except Exception as e:
        logger.error("Failed to upload logs to fpaste: %s", e)
        out["url"] = None
        out["error"] = str(e)


class DebugPlugin(HHDPlugin):
    def __init__(self) -> None:
        self.name = f"debug"
        self.priority = 80
        self.log = "DDBG"
        self.emit = None
        self.t = None
        self.error = None
        self.fpaste_t = None
        self.fpaste_data = None

    def settings(self) -> HHDSettings:
        sets = {"debug": load_relative_yaml("settings.yml")}
        if not USES_BETA:
            del sets["debug"]["dev"]["children"]["hhd_dev_exit"]
        return sets

    def open(
        self,
        emit,
        context: Context,
    ):
        self.emit = emit

    def update(self, conf: Config):
        self._hhd_dev(conf)
        self._fpaste(conf)

    def _fpaste(self, conf: Config):
        fpaste = conf.get_action("debug.reports.submit")

        if self.fpaste_t:
            if self.fpaste_t.is_alive():
                return
            self.fpaste_t.join()
            self.fpaste_t = None
            if self.fpaste_data and self.fpaste_data.get("url", None):
                conf["debug.reports.url"] = self.fpaste_data["url"]
            elif self.fpaste_data and self.fpaste_data.get("error", None):
                conf["debug.reports.error"] = self.fpaste_data["error"]
            conf["debug.reports.progress"] = None
        elif fpaste:
            conf["debug.reports.progress"] = {
                "text": _("Uploading log to fpaste..."),
                "value": None,
                "unit": None,
            }
            conf["debug.reports.url"] = None
            conf["debug.reports.error"] = None
            self.fpaste_data = {}
            self.fpaste_t = Thread(
                target=upload_log,
                args=(conf.get("debug.reports.boot", "current"), self.fpaste_data),
            )
            self.fpaste_t.start()

    def _hhd_dev(self, conf: Config):
        hhd_dev = conf.get_action("debug.dev.hhd_dev")

        if USES_BETA and self.emit and conf.get_action("debug.dev.hhd_dev_exit"):
            conf["debug.dev.progress"] = {
                "text": _("Shutting down..."),
                "value": None,
                "unit": None,
            }
            self.emit({"type": "special", "event": "shutdown_dev"})

        if self.t:
            if self.t.is_alive():
                return
            self.t.join()
            self.t = None
            if self.error and self.error.is_set():
                conf["debug.dev.progress"] = None
                conf["debug.dev.error"] = _("Failed to download Handheld Daemon Beta.")
            elif self.emit:
                self.emit({"type": "special", "event": "restart_dev"})

        if hhd_dev:
            conf["debug.dev.progress"] = {
                "text": _("Downloading Beta and Restarting..."),
                "value": None,
                "unit": None,
            }
            self.error = Event()
            self.t = Thread(target=prepare_hhd_dev, args=(self.error,))
            self.t.start()

    def close(self):
        if self.t:
            self.t.join()
            self.t = None
        if self.fpaste_t:
            self.fpaste_t.join()
            self.fpaste_t = None


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    if not BUGREPORTS_ENABLED:
        return []

    return [DebugPlugin()]



from typing import Literal, NamedTuple, Sequence, TypedDict


class PowerButtonConfig(NamedTuple):
    device: str
    prod_name: str
    type: Literal["hold_emitted", "hold_isa", "only_press", "disabled"] = "hold_isa"
    phys: Sequence[str] = ["LNXPWRBN", "PNP0C0C"]
    hold_phys: Sequence[str] = ["phys-hhd-powerbutton", "isa0060"]
    hold_grab: bool = False
    hold_code: int = 125  # left meta
    unsupported: bool = False


# POWER_BUTTON_NAMES = ["Power Button"]
# POWER_BUTTON_PHYS = ["LNXPWRBN", "PNP0C0C"]

PBC = PowerButtonConfig

SUPPORTED_DEVICES: Sequence[PowerButtonConfig] = [
    PBC("Legion Go", "83E1"),
    PBC("Legion Go S Z2 Go", "83L3"),
    PBC("Legion Go S Z1E", "83N6"),
    PBC("Legion Go S", "83Q2"),
    PBC("Legion Go S", "83Q3"),
    PBC("ROG Ally", "ROG Ally RC71L_Action"),
    PBC("ROG Ally", "ROG Ally RC71L_RC71L"),
    PBC("ROG Ally", "ROG Ally RC71L"),
    PBC("ROG Ally X", "ROG Ally X RC72LA"),
    PBC("GPT Win 4", "G1618-04"),
    PBC("GPD Win Mini", "G1617-01"),
    PBC("GPD Win Mini", "G1617-02"),
    PBC("GPD Win Max 2", "G1619-04"),
    PBC("GPD Win Max 2", "G1619-05"),
    PBC("OrangePi G1621-02/G1621-02", "G1621-02"),
    PBC("OrangePi NEO-01/NEO-01", "NEO-01"),
    # breaks volume buttons, use the valve original script and hope steam inhibits systemd
    PBC(
        "Steam Deck LCD",
        "Jupiter",
        type="disabled",
    ),
    PBC(
        "Steam Deck OLED",
        "Galileo",
        type="disabled",
    ),
    PBC(
        "AOKZOE A1",
        "AOKZOE A1 AR07",
        type="only_press",
        phys=["LNXPWRBN", "PNP0C0C"],
    ),
    PBC(
        "AOKZOE A1 Pro",
        "AOKZOE A1 Pro",
        type="only_press",
        phys=["LNXPWRBN", "PNP0C0C"],
    ),
    PBC(
        "ONEXPLAYER Mini Pro",
        "ONEXPLAYER Mini Pro",
        type="only_press",
        phys=["LNXPWRBN", "PNP0C0C"],
    ),
    PBC(
        "TECNO (Displayless)",
        "Pocket Go",
        type="only_press",
    ),
    PBC(
        "MSI Claw 8",
        "Claw 8 AI+ A2VM",
        type="only_press",
        phys=["LNXPWRBN"],
    ),
    PBC(
        "MSI Claw A8",
        "Claw A8 BZ2EM",
        type="only_press",
    ),
]


def get_config() -> PowerButtonConfig:
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        prod = f.read().strip()

    try:
        with open("/sys/devices/virtual/dmi/id/sys_vendor") as f:
            sys = f.read().strip()
    except Exception:
        sys = None

    for d in SUPPORTED_DEVICES:
        if d.prod_name in prod:
            return d

    if "ONEXPLAYER" in prod or "AOKZOE" in prod:
        return PBC(prod, prod, type="only_press")

    if sys == "AYA" or sys == "AYANEO" or sys == "AYN":
        # TODO: Fix isa handling to only work when only shift is active
        return PBC(prod, prod, type="only_press")

    return PBC("uknown", "NA", "only_press", unsupported=True)


# Legion go
# At device with phys=isa0060/serio0/input0
#
# event at 1702329391.831152, code 04, type 04, val 219
# event at 1702329391.831152, code 125, type 01, val 01
# event at 1702329391.831152, code 00, type 00, val 00
# event at 1702329391.834208, code 04, type 04, val 103
# event at 1702329391.834208, code 00, type 00, val 00

# event at 1702329392.152298, code 04, type 04, val 219
# event at 1702329392.152298, code 125, type 01, val 00
# event at 1702329392.152298, code 00, type 00, val 00
# event at 1702329392.156347, code 04, type 04, val 103
# event at 1702329392.156347, code 00, type 00, val 00



import subprocess
import os
import logging
import threading

from .ioctl import EVIOCREVOKEALL, JSIOCREVOKEALL
from fcntl import ioctl

logger = logging.getLogger(__name__)

ENHANCED_HIDING = os.environ.get("HHD_EVIOC_IOCTL", "0") == "1"
HIDE_ALL = os.environ.get("HHD_HIDE_ALL", "0") == "1"

_hidden = []

def get_device_info(devpath: str):
    syspath = None
    for line in subprocess.run(
        ["udevadm", "info", devpath], capture_output=True
    ).stdout.splitlines():
        if line.startswith(b"P: "):
            return line[3:].decode()

    return syspath


def get_gamepad_name(syspath: str):
    parts = syspath.split("/")
    if len(parts) < 3:
        return None
    input_dev = parts[-2]
    if not input_dev.startswith("input") or input_dev == "input":
        return None
    return input_dev


def get_parent_sysfs(syspath: str):
    return syspath[: syspath.rindex("/")]
    # return syspath.split("/input/")[0]


_reload_thread = None


def _reload_children_worker(parent: str):
    stat = subprocess.run(
        ["udevadm", "control", "--reload-rules"],
        capture_output=True,
    )
    if stat.returncode:
        return False
    for action in ["remove", "add"]:
        stat = subprocess.run(
            ["udevadm", "trigger", "--action", action, "-b", parent],
            capture_output=True,
        )
        if stat.returncode:
            return False
    return True


def reload_children(parent: str):
    global _reload_thread

    if _reload_thread:
        _reload_thread.join()
        _reload_thread = None

    _reload_thread = threading.Thread(target=_reload_children_worker, args=(parent,))
    _reload_thread.start()


def hide_gamepad(devpath: str, vid: int, pid: int) -> str | None:
    syspath = get_device_info(devpath)
    if not syspath:
        return None
    input_dev = get_gamepad_name(syspath)
    parent = get_parent_sysfs(syspath)
    if not input_dev or not parent:
        return None

    if HIDE_ALL:
        # Hide all devices with the same vid pid
        root = f"{vid:04x}-{pid:04x}"
        # Certain devices emulate a USB Xbox controller. So match the USB bus
        # to hopefully not affect bluetooth devices.
        extra = 'ENV{ID_BUS}=="usb"'
    else:
        root = input_dev
        extra = f'KERNELS=="{input_dev}", '

    out_fn = f"/run/udev/rules.d/95-hhd-devhide-{root}.rules"
    if os.path.exists(out_fn):
        # Skip hiding controller on reloads
        return input_dev

    rule = f"""\
# Hides device gamepad devices stemming from {input_dev}
# Managed by HHD, this file will be autoremoved during configuration changes.
SUBSYSTEMS=="input", {extra}ATTRS{{id/vendor}}=="{vid:04x}", ATTRS{{id/product}}=="{pid:04x}", GOTO="hhd_valid"
GOTO="hhd_end"
LABEL="hhd_valid"
KERNEL=="js[0-9]*|event[0-9]*", SUBSYSTEM=="input", MODE="000", GROUP="root", TAG-="uaccess", RUN+="/bin/chmod 000 /dev/input/%k"
LABEL="hhd_end"
"""

    #     # Hide usb xinput, be very careful to only match that usb
    #     if "/" in parent:
    #         usb_root = parent[parent.rindex("/") + 1 :]
    #         if re.match(r"\d-+\d+", usb_root) or re.match(r"\d+-\d+:\d+\.\d+", usb_root):
    #             rule += f"""
    # # Hides the Xinput/Hidraw input node so that certain games that access it directly.
    # SUBSYSTEMS=="usb", ATTRS{{idVendor}}=="{vid:04x}", ATTRS{{idProduct}}=="{pid:04x}",\
    #  KERNEL=="{usb_root}", TAG-="uaccess", GROUP="root", MODE="000"
    # """

    try:
        # Add udev rules to strip the device perms from the system
        os.makedirs("/run/udev/rules.d/", exist_ok=True)
        with open(out_fn, "w") as f:
            f.write(rule)
        # Reload the rules for that device to make it owned by root
        reload_children(parent)
        _hidden.append(parent)

        # Use flag until further testing
        if not ENHANCED_HIDING:
            return input_dev

        # Now that only we can access the device, revoke open fds
        # Custom kernel feature. NOOP if it fails.
        try:
            for fn in os.listdir("/sys/" + parent):
                if fn.startswith("event"):
                    ioc = EVIOCREVOKEALL
                elif fn.startswith("js"):
                    ioc = JSIOCREVOKEALL
                else:
                    continue

                fd = None
                try:
                    dev = os.path.join("/dev/input", fn)
                    fd = os.open(dev, os.O_RDONLY)
                    ioctl(fd, ioc, 0)
                    logger.info(f"Revoked access to device '{dev}'.")
                finally:
                    if fd:
                        os.close(fd)
        except Exception as e:
            logger.exception(
                f"Failed to run EV/JSIOCREVOKEALL. Games may remember the controller. Error:\n{e}"
            )

        return input_dev
    except Exception:
        return None


def unhide_gamepad(devpath: str, root: str | None = None):
    if HIDE_ALL:
        # Do not unhide device to be ready when the next one shows up
        return False

    try:
        # Remove file before searching for device
        if root is not None:
            os.remove(f"/run/udev/rules.d/95-hhd-devhide-{root}.rules")
    except Exception:
        return False

    syspath = get_device_info(devpath)
    if not syspath:
        return False
    input_dev = get_gamepad_name(syspath)
    parent = get_parent_sysfs(devpath)
    if not input_dev or not parent:
        return False

    if parent in _hidden:
        _hidden.remove(parent)

    try:
        if root is None:
            os.remove(f"/run/udev/rules.d/95-hhd-devhide-{input_dev}.rules")
        return reload_children(parent)
    except Exception:
        return False


def unhide_all():
    removed = False
    try:
        for rule in os.listdir("/run/udev/rules.d/"):
            if rule.startswith("95-hhd-devhide"):
                os.remove(os.path.join("/run/udev/rules.d/", rule))
                logger.info(f"Removed rule '{rule}'.")
                removed = True
    except Exception:
        pass

    if not removed:
        return True

    # We have to reload affected devices if we removed rules
    for parent in _hidden:
        reload_children(parent)
    _hidden.clear()


import logging
import os
from threading import Event, Thread
from typing import Any, Sequence

from hhd.plugins import (
    Config,
    Context,
    Emitter,
    HHDPlugin,
    get_gyro_config,
    get_outputs_config,
    load_relative_yaml,
)
from hhd.plugins.settings import HHDSettings

from .const import CONFS, DEFAULT_MAPPINGS, get_default_config

logger = logging.getLogger(__name__)


class GenericControllersPlugin(HHDPlugin):
    name = "onexplayer"
    priority = 18
    log = "oxpc"

    def __init__(self, dmi: str, dconf: dict) -> None:
        self.t = None
        self.should_exit = None
        self.updated = Event()
        self.started = False
        self.t = None

        self.dmi = dmi
        self.dconf = dconf
        self.name = f"onexplayer@'{dconf.get('name', 'ukn')}'"

    def open(
        self,
        emit: Emitter,
        context: Context,
    ):
        self.emit = emit
        self.context = context
        self.prev = None

        # Use the oxp-platform driver if available
        turbo = False
        if self.dconf.get("turbo", True) and os.path.exists(
            "/sys/devices/platform/oxp-platform/tt_toggle"
        ):
            try:
                with open("/sys/devices/platform/oxp-platform/tt_toggle", "w") as f:
                    f.write("1")
                logger.info(f"Turbo button takeover enabled")
                turbo = True

                if os.path.exists("/sys/devices/platform/oxp-platform/tt_led"):
                    with open("/sys/devices/platform/oxp-platform/tt_led", "w") as f:
                        f.write("0")
            except Exception:
                logger.warning(
                    f"Turbo takeover failed. Ensure you have the latest oxp-sensors driver installed."
                )
        self.turbo = turbo

    def notify(self, events: Sequence):
        if not self.turbo:
            return

        woke = False
        for ev in events:
            if ev["type"] == "special" and ev.get("event", None) == "wakeup":
                woke = True

        if not woke:
            return

        # We need to reset after hibernation
        try:
            logger.info(f"Turbo button takeover enabled")
            with open("/sys/devices/platform/oxp-platform/tt_toggle", "w") as f:
                f.write("1")

            if os.path.exists("/sys/devices/platform/oxp-platform/tt_led"):
                with open("/sys/devices/platform/oxp-platform/tt_led", "w") as f:
                    f.write("0")
        except Exception:
            logger.warning(
                f"Turbo takeover failed. Ensure you have the latest oxp-sensors driver installed."
            )

    def settings(self) -> HHDSettings:
        base = {"controllers": {"oxp": load_relative_yaml("controllers.yml")}}
        base["controllers"]["oxp"]["children"]["controller_mode"].update(
            get_outputs_config(
                can_disable=True,
                has_leds=self.dconf.get("rgb", True),
                start_disabled=self.dconf.get("untested", False),
                extra_buttons=self.dconf.get("extra_buttons", "dual"),
            )
        )

        base["controllers"]["oxp"]["children"]["imu_axis"] = get_gyro_config(
            self.dconf.get("mapping", DEFAULT_MAPPINGS)
        )

        if not self.dconf.get("x1", False):
            del base["controllers"]["oxp"]["children"]["volume_reverse"]
            # Maybe it is helpful for OneXFly users
            # del base["controllers"]["oxp"]["children"]["swap_face"]

        if not self.turbo:
            del base["controllers"]["oxp"]["children"]["extra_buttons"]
            del base["controllers"]["oxp"]["children"]["turbo_reboots"]
        elif self.dconf.get("g1"):
            # No Keyboard button means turbo needs to do combo
            del base["controllers"]["oxp"]["children"]["extra_buttons"]

        return base

    def update(self, conf: Config):
        new_conf = conf["controllers.oxp"]
        if new_conf == self.prev:
            return
        if self.prev is None:
            self.prev = new_conf
        else:
            self.prev.update(new_conf.conf)

        self.updated.set()
        self.start(self.prev)

    def start(self, conf):
        from .base import plugin_run

        if self.started:
            return
        self.started = True

        self.close()
        self.should_exit = Event()
        self.t = Thread(
            target=plugin_run,
            args=(
                conf,
                self.emit,
                self.context,
                self.should_exit,
                self.updated,
                self.dconf,
                self.turbo,
            ),
        )
        self.t.start()

    def close(self):
        if not self.should_exit or not self.t:
            return
        self.should_exit.set()
        self.t.join()
        self.should_exit = None
        self.t = None

        if self.turbo:
            # Disable turbo button takeover
            try:
                with open("/sys/devices/platform/oxp-platform/tt_toggle", "w") as f:
                    f.write("0")
            except Exception:
                pass


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    # Match just product name
    # if a device exists here its officially supported
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        dmi = f.read().strip()

    dconf = CONFS.get(dmi, None)
    if dconf:
        return [GenericControllersPlugin(dmi, dconf)]

    # Begin hw agnostic dmi match
    if "ONEXPLAYER" in dmi:
        return [GenericControllersPlugin(dmi, get_default_config(dmi, "ONEXPLAYER"))]

    return []



import logging
import re
from typing import Any, Literal, NamedTuple, Protocol, Sequence

from hhd.controller import (
    Axis,
    Button,
    Configuration,
    Consumer,
    Event,
    Producer,
    can_read,
)
from hhd.controller.base import Event
from hhd.controller.lib.common import (
    AM,
    BM,
    CM,
    decode_axis,
    decode_config,
    get_button,
    hexify,
    matches_patterns,
)
from hhd.controller.lib.hid import MAX_REPORT_SIZE, Device, enumerate_unique

logger = logging.getLogger(__name__)


class EventCallback(Protocol):
    def __call__(self, dev: Device, events: Sequence[Event]) -> Any:
        pass


class GenericGamepadHidraw(Producer, Consumer):
    def __init__(
        self,
        vid: Sequence[int] = [],
        pid: Sequence[int] = [],
        manufacturer: Sequence[str | re.Pattern] = [],
        product: Sequence[str | re.Pattern] = [],
        usage_page: Sequence[int] = [],
        usage: Sequence[int] = [],
        application: Sequence[int] = [],
        interface: int | None = None,
        btn_map: dict[int | None, dict[Button, BM]] = {},
        axis_map: dict[int | None, dict[Axis, AM]] = {},
        config_map: dict[int | None, dict[Configuration, CM]] = {},
        callback: EventCallback | None = None,
        report_size: int = MAX_REPORT_SIZE,
        required: bool = True,
        lossless: bool = True,
    ) -> None:
        self.vid = vid
        self.pid = pid
        self.manufacturer = manufacturer
        self.product = product
        self.usage_page = usage_page
        self.usage = usage
        self.interface = interface
        self.info = None
        self.report_size = report_size

        self.application = application
        self.btn_map = btn_map
        self.axis_map = axis_map
        self.config_map = config_map
        self.callback = callback
        self.required = required
        self.lossless = lossless

        self.path = None
        self.dev: Device | None = None
        self.fd = 0

        self.report = None

    def open(self) -> Sequence[int]:
        for d in enumerate_unique():
            if not matches_patterns(d["vendor_id"], self.vid):
                continue
            if not matches_patterns(d["product_id"], self.pid):
                continue
            if not matches_patterns(d["manufacturer_string"], self.manufacturer):
                continue
            if not matches_patterns(d["product_string"], self.product):
                continue
            if not matches_patterns(d["usage_page"], self.usage_page):
                continue
            if not matches_patterns(d["usage"], self.usage):
                continue
            if not matches_patterns(
                (d["usage_page"] << 16) + d["usage"], self.application
            ):
                continue
            if (
                self.interface is not None
                and d.get("interface_number", None) != self.interface
            ):
                continue
            self.path = d["path"]
            self.dev = Device(path=self.path)
            self.info = d
            self.fd = self.dev.fd
            logger.info(
                f"Found device {hexify(d['vendor_id'])}:{hexify(d['product_id'])}:\n"
                + f"'{d['manufacturer_string']}': '{d['product_string']}' at {d['path']}"
            )
            self.report = None
            self.prev_btn = {}
            self.prev_axis = {}
            self.prev_config = {}
            return [self.fd]

        err = f"Device with the following not found:\n"
        if self.vid:
            err += f"Vendor ID: {hexify(self.vid)}\n"
        if self.pid:
            err += f"Product ID: {hexify(self.pid)}\n"
        if self.manufacturer:
            err += f"Manufacturer: {self.manufacturer}\n"
        if self.product:
            err += f"Product: {self.product}\n"
        if self.usage_page:
            err += f"Usage Page: {hexify(self.usage_page)}\n"
        if self.usage:
            err += f"Usage: {hexify(self.usage)}\n"
        if self.application:
            err += f"Application: {hexify(self.application)}\n"
        logger.error(err)
        if self.required:
            raise RuntimeError()
        return []

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        # If we can not read return
        if not self.fd or self.fd not in fds or not self.dev:
            return []
        rep = None

        if self.lossless:
            # Keep all events
            rep = self.dev.read(self.report_size)
        else:
            # Throw away stale events
            while can_read(self.fd):
                rep = self.dev.read(self.report_size)

        # If we could not read (?) return
        if not rep:
            return []

        # If the report is the same as the previous one, return
        if self.report and self.report == rep:
            return []
        self.report = rep
        rep_id = rep[2] if len(rep) > 2 else None

        # Allow for devices with NULL reports
        if None in self.btn_map or None in self.axis_map:
            rep_id = None

        # Decode buttons
        out: list[Event] = []
        if rep_id in self.btn_map:
            for btn, map in self.btn_map[rep_id].items():
                val = get_button(rep, map)
                if btn in self.prev_btn and self.prev_btn[btn] == val:
                    continue
                self.prev_btn[btn] = val
                out.append({"type": "button", "code": btn, "value": val})

        # Decode Axis
        if rep_id in self.axis_map:
            for ax, map in self.axis_map[rep_id].items():
                val = decode_axis(rep, map)
                if ax in self.prev_axis and self.prev_axis[ax] == val:
                    continue
                self.prev_axis[ax] = val
                out.append({"type": "axis", "code": ax, "value": val})

        # Decode
        if rep_id in self.config_map:
            for cnf, map in self.config_map[rep_id].items():
                val = decode_config(rep, map)
                if cnf in self.prev_config and self.prev_config[cnf] == val:
                    continue
                self.prev_config[cnf] = val
                out.append({"type": "configuration", "code": cnf, "value": val})
        return out

    def consume(self, events: Sequence[Event]):
        if self.callback and self.dev:
            self.callback(self.dev, events)

    def close(self, exit: bool) -> bool:
        if self.dev:
            self.dev.close()
            self.dev = None
        return True


__all__ = ["GenericGamepadHidraw", "BM", "AM"]



from threading import Condition, Thread, Event
import time
import random

CACHE_TIMEOUT = 20
UPDATE_FREQ = 25
UPDATE_T = 1 / UPDATE_FREQ


class ControllerCache:
    def __init__(self, cache_timeout=CACHE_TIMEOUT) -> None:
        self._t = None
        self._cond = Condition()
        self._cached = None
        self._should_exit = Event()
        self.cache_timeout = cache_timeout

    def _close_cached(self):
        with self._cond:
            delay = 0
            while delay < self.cache_timeout and not self._should_exit.is_set():
                self._cond.wait(UPDATE_T)
                delay += UPDATE_T
                if self._cached:
                    # Send fake events to keep everyone happy
                    # Both steam and kernel
                    self._cached.produce([self._cached.fd])
                    # Bridge timestamps to prevent jump after the error
                    # when disconnecting and reconnecting the controller
                    # provided the imu timestamp does not break after reloading
                    # Use .95 to prevent racing ahead of the IMU ts
                    if hasattr(self._cached, "last_imu_ts"):
                        ctime = getattr(self._cached, "last_imu_ts") + int(
                            delay * 0.95 * 1e9
                        )
                    else:
                        ctime = int(delay * 1e9)

                    # Send a lot of noise to the accel value to avoid
                    # steam recalibrating. Only RPCS3 and dolphin might
                    # have an issue with this.
                    accel = random.random() * 10
                    self._cached.consume(
                        [
                            {"type": "axis", "code": "left_imu_ts", "value": ctime},
                            {"type": "axis", "code": "right_imu_ts", "value": ctime},
                            {"type": "axis", "code": "imu_ts", "value": ctime},
                            {"type": "axis", "code": "accel_x", "value": accel},
                            {"type": "axis", "code": "left_accel_x", "value": accel},
                            {"type": "axis", "code": "right_accel_x", "value": accel},
                        ]
                    )
                else:
                    # Exit if cached became null during sleep
                    break
            if self._cached:
                self._cached.close(True, in_cache=True)
                self._cached = None

    def add(self, c):
        tmp = None
        with self._cond:
            if self._t:
                self._should_exit.set()
                self._cond.notify_all()
                tmp = self._t
                self._t = None
        if tmp:
            tmp.join()

        with self._cond:
            self._cached = c
            self._should_exit.clear()
            self._t = Thread(target=self._close_cached)
            self._t.start()

    def get(self):
        with self._cond:
            tmp = self._cached
            self._cached = None
            self._should_exit.set()
            self._cond.notify_all()
            tmp2 = self._t
            self._t = None
        if tmp2:
            tmp2.join()
        return tmp

    def close(self):
        with self._cond:
            if self._cached:
                self._cached.close(True, in_cache=True)
                self._cached = None
            self._should_exit.set()
            self._cond.notify_all()



import logging
from typing import Any, Literal, Mapping, Sequence

import os
from ..controller.base import Consumer, Producer, RgbMode, RgbSettings, RgbZones
from ..controller.virtual.dualsense import Dualsense, TouchpadCorrectionType
from ..controller.virtual.sd import SteamdeckController
from ..controller.virtual.uinput import (
    CONTROLLER_THEMES,
    GAMEPAD_BUTTON_MAP,
    HHD_PID_TOUCHPAD,
    HORIPAD_STEAM_BUTTON_MAP,
    MOTION_AXIS_MAP,
    MOTION_AXIS_MAP_FLIP_Z,
    MOTION_CAPABILITIES,
    MOTION_INPUT_PROPS,
    TOUCHPAD_AXIS_MAP,
    TOUCHPAD_BUTTON_MAP,
    TOUCHPAD_CAPABILITIES,
    XBOX_ELITE_BUTTON_MAP,
    UInputDevice,
)
from .plugin import is_steam_gamepad_running, open_steam_kbd
from .utils import load_relative_yaml

logger = logging.getLogger(__name__)

HORI_ENABLED = os.environ.get("HHD_HORI_STEAM", "0") == "1"


def get_outputs(
    conf,
    touch_conf,
    motion: bool = False,
    *,
    controller_id: int = 0,
    emit=None,
    dual_motion: bool = False,
    rgb_modes: Mapping[RgbMode, Sequence[RgbSettings]] | None = None,
    rgb_zones: RgbZones = "mono",
    rgb_resets_on_ac: bool = False,
    controller_disabled: bool = False,
    touchpad_enable: Literal["disabled", "gamemode", "always"] | None = None,
) -> tuple[Sequence[Producer], Sequence[Consumer], Mapping[str, Any]]:
    producers = []
    consumers = []
    nintendo_qam = False

    controller = conf["mode"].to(str)
    desktop_disable = False
    if touch_conf is not None:
        touchpad = touch_conf["mode"].to(str)
        correction = touch_conf["controller.correction"].to(TouchpadCorrectionType)
        if touchpad in ("emulation", "controller"):
            desktop_disable = touch_conf[touchpad]["desktop_disable"].to(bool)
    elif touchpad_enable:
        touchpad = "disabled" if touchpad_enable == "disabled" else "controller"
        correction = "legos"  # todo: make generic
        desktop_disable = touchpad_enable == "gamemode"
    else:
        touchpad = "controller"
        correction = "stretch"

    # Run steam check for touchpad
    steam_check = (
        is_steam_gamepad_running(emit.ctx) if emit and desktop_disable else None
    )
    match steam_check:
        case True:
            logger.info("Gamepadui active. Activating touchpad emulation.")
        case False:
            logger.info("Gamepadui closed. Disabling touchpad emulation.")

    has_qam = False
    uses_touch = False
    uses_leds = False
    noob_mode = False
    flip_z = False

    if controller_disabled:
        controller = "hidden"

    if controller == "hori_steam" and conf.get("hori_steam.handheld_mode", False):
        controller = "sd"

    match controller:
        case "hidden":
            # NOOP
            UInputDevice.close_cached()
            Dualsense.close_cached()
            SteamdeckController.close_cached()
            motion = False
            noob_mode = conf.get("hidden.noob_mode", False)
        case "dualsense":
            UInputDevice.close_cached()
            SteamdeckController.close_cached()
            flip_z = conf["dualsense.flip_z"].to(bool)
            uses_touch = touchpad == "controller" and steam_check is not False
            uses_leds = conf.get("dualsense.led_support", False)
            paddles_as = conf.get("dualsense.paddles_as", "noob")
            noob_mode = paddles_as in ("noob", "both")
            edge_mode = False

            if paddles_as == "both":
                paddles_to_clicks = "bottom"
            elif paddles_as == "touchpad":
                paddles_to_clicks = "top"
            elif paddles_as == "steam_input":
                edge_mode = True
                paddles_to_clicks = "disabled"
            else:
                paddles_to_clicks = "disabled"

            d = Dualsense(
                touchpad_method=correction,
                edge_mode=edge_mode,
                use_bluetooth=conf["dualsense.bluetooth_mode"].to(bool),
                enable_touchpad=uses_touch,
                enable_rgb=uses_leds,
                fake_timestamps=not motion,
                sync_gyro=conf["dualsense.sync_gyro"].to(bool) and motion,
                paddles_to_clicks=paddles_to_clicks,
                flip_z=flip_z,
                controller_id=controller_id | (0xF0 if edge_mode else 0),
                cache=True,
            )
            producers.append(d)
            consumers.append(d)
        case "sd":
            UInputDevice.close_cached()
            Dualsense.close_cached()
            uses_touch = touchpad == "controller" and steam_check is not False
            d = SteamdeckController(
                name="Steam Controller (HHD)",
                pid=0x12FF,
                touchpad=uses_touch,
                sync_gyro=conf.get("hori_steam.sync_gyro", True),
            )
            producers.append(d)
            consumers.append(d)
            noob_mode = conf.get("hori_steam.noob_mode", False)
            has_qam = True
        case "uinput" | "xbox_elite" | "joycon_pair" | "hori_steam":
            Dualsense.close_cached()
            SteamdeckController.close_cached()
            version = 1
            sync_gyro = False
            paddles_as = conf.get("uinput.paddles_as", "noob")
            if controller == "joycon_pair":
                theme = "joycon_pair"
                nintendo_qam = conf["joycon_pair.nintendo_qam"].to(bool)
                button_map = GAMEPAD_BUTTON_MAP
                bus = 0x06
                version = 0
            elif controller == "hori_steam":
                theme = "hori_steam"
                noob_mode = conf.get("hori_steam.noob_mode", False)
                flip_z = conf["hori_steam.flip_z"].to(bool)
                button_map = HORIPAD_STEAM_BUTTON_MAP
                bus = 0x06
                version = 0
                sync_gyro = conf.get("hori_steam.sync_gyro", True)
                has_qam = True
            elif controller == "xbox_elite" or (
                controller == "uinput" and paddles_as == "steam_input"
            ):
                theme = "xbox_one_elite"
                button_map = XBOX_ELITE_BUTTON_MAP
                bus = 0x03
            else:
                noob_mode = paddles_as == "noob"
                # theme = conf.get("uinput.theme", "hhd")
                theme = "hhd"
                nintendo_qam = conf["uinput.nintendo_qam"].to(bool)
                # flip_z = conf["uinput.flip_z"].to(bool)
                flip_z = False
                button_map = GAMEPAD_BUTTON_MAP
                bus = 0x03 if theme == "hhd" else 0x06
            vid, pid, name = CONTROLLER_THEMES[theme]
            addr = "phys-hhd-main"
            if controller_id:
                addr = f"phys-hhd-{controller_id:02d}"
            d = UInputDevice(
                name=name,
                vid=vid,
                pid=pid,
                phys=addr,
                uniq=addr,
                btn_map=button_map,
                bus=bus,
                version=version,
                cache=True,
                sync_gyro=sync_gyro and motion,
            )
            producers.append(d)
            consumers.append(d)
            # Deactivate motion if using an xbox theme
            motion = (theme != "hhd" and "xbox" not in theme) and motion
            if motion:
                d = UInputDevice(
                    name=f"{name} Motion Sensors",
                    vid=vid,
                    pid=pid,
                    phys=addr,
                    uniq=addr,
                    bus=bus,
                    version=version,
                    capabilities=MOTION_CAPABILITIES,
                    btn_map={},
                    axis_map=(MOTION_AXIS_MAP_FLIP_Z if flip_z else MOTION_AXIS_MAP),
                    output_imu_timestamps=True,
                    input_props=MOTION_INPUT_PROPS,
                    ignore_cmds=True,
                    cache=True,
                    motions_device=True,
                )
                producers.append(d)
                consumers.append(d)
        case _:
            raise RuntimeError(f"Invalid controller type: '{controller}'.")

    dual_motion = motion and dual_motion
    if dual_motion:
        d = Dualsense(
            edge_mode=False,
            use_bluetooth=True,
            enable_touchpad=False,
            enable_rgb=False,
            fake_timestamps=False,
            sync_gyro=True,
            paddles_to_clicks="disabled",
            flip_z=flip_z,
            controller_id=5,
            cache=True,
            left_motion=True,
        )
        producers.append(d)
        consumers.append(d)

    if touchpad == "emulation" and steam_check is not False and not controller_disabled:
        d = UInputDevice(
            name="Handheld Daemon Touchpad",
            phys="phys-hhd-main",
            capabilities=TOUCHPAD_CAPABILITIES,
            pid=HHD_PID_TOUCHPAD,
            btn_map=TOUCHPAD_BUTTON_MAP,
            axis_map=TOUCHPAD_AXIS_MAP,
            output_timestamps=True,
            ignore_cmds=True,
        )
        producers.append(d)
        consumers.append(d)
        uses_touch = True

    return (
        producers,
        consumers,
        {
            "uses_touch": uses_touch,
            "rgb_used": uses_leds,
            "rgb_modes": rgb_modes,
            "rgb_zones": rgb_zones,
            "is_dual": False,
            "steam_check": steam_check,
            "steam_check_fn": lambda: emit and is_steam_gamepad_running(emit.ctx),
            "steam_kbd": lambda open: open_steam_kbd(emit, open),
            "nintendo_qam": nintendo_qam,
            "uses_motion": motion,
            "uses_dual_motion": dual_motion,
            "noob_mode": noob_mode,
            "has_qam": has_qam,
            "supports_qam": not controller_disabled and controller != "hidden",
            "rgb_resets_on_ac": rgb_resets_on_ac,
        },
    )


def get_outputs_config(
    can_disable: bool = False,
    has_leds: bool = True,
    start_disabled: bool = False,
    default_device: str | None = None,
    extra_buttons: Literal["none", "dual", "quad"] = "dual",
):
    s = load_relative_yaml("outputs.yml")
    try:
        if not can_disable:
            del s["modes"]["disabled"]
        if not has_leds:
            del s["modes"]["dualsense"]["children"]["led_support"]

        if extra_buttons == "none":
            del s["modes"]["dualsense"]["children"]["paddles_as"]
            del s["modes"]["uinput"]["children"]["paddles_as"]
            del s["modes"]["hidden"]["children"]["noob_mode"]
            del s["modes"]["xbox_elite"]
        elif extra_buttons == "dual":
            del s["modes"]["dualsense"]["children"]["paddles_as"]["options"]["both"]

        if HORI_ENABLED:
            # Replace xbox elite with hori
            try:
                del s["modes"]["xbox_elite"]
            except Exception:
                pass
        else:
            del s["modes"]["hori_steam"]

        # Set xbox as default for now
        s["default"] = "uinput"

        # if default_device:
        #     s["default"] = default_device
        if start_disabled:
            s["default"] = "disabled"
    except Exception as e:
        logger.exception(f"Error fixing outputs:\n{e}")
    return s


def get_limits_config(defaults: dict[str, int] = {}):
    s = load_relative_yaml("limits.yml")
    lims = s["modes"]["manual"]["children"]

    lims["ls_min"]["default"] = defaults.get("s_min", 0)
    lims["ls_max"]["default"] = defaults.get("s_max", 95)
    lims["rs_min"]["default"] = defaults.get("s_min", 0)
    lims["rs_max"]["default"] = defaults.get("s_max", 95)

    lims["rt_min"]["default"] = defaults.get("t_min", 0)
    lims["rt_max"]["default"] = defaults.get("t_max", 95)
    lims["lt_min"]["default"] = defaults.get("t_min", 0)
    lims["lt_max"]["default"] = defaults.get("t_max", 95)

    return s


def get_limits(conf, defaults={}):
    if conf["mode"].to(str) != "manual":
        return defaults

    kconf = conf["manual"]
    for set in ("ls", "rs", "lt", "rt"):
        if kconf[f"{set}_min"].to(int) > kconf[f"{set}_max"].to(int):
            kconf[f"{set}_min"] = kconf[f"{set}_max"].to(int)
    return kconf.to(dict)


def fix_limits(conf, prefix: str, defaults: dict[str, int] = {}):
    if conf[f"{prefix}.mode"].to(str) != "manual":
        return {}

    # Make sure min < max
    for set in ("ls", "rs", "lt", "rt"):
        if conf[f"{prefix}.manual.{set}_min"].to(int) > conf[
            f"{prefix}.manual.{set}_max"
        ].to(int):
            conf[f"{prefix}.manual.{set}_min"] = conf[f"{prefix}.manual.{set}_max"].to(
                int
            )

    # Allow reseting limits
    if conf[f"{prefix}.manual.reset"].to(bool):
        for set in ("l", "r"):
            for comp in ("t", "s"):
                conf[f"{prefix}.manual.{set}{comp}_min"] = defaults.get(
                    f"{comp}_min", 0
                )
                conf[f"{prefix}.manual.{set}{comp}_max"] = defaults.get(
                    f"{comp}_max", 95
                )
        conf[f"{prefix}.manual.reset"] = False



import logging
import os
import time
from threading import Event as TEvent
from threading import Thread
from typing import Any, Sequence

from hhd.controller import Consumer
from hhd.controller.base import Event, RgbLedEvent

LED_BASE = "/sys/class/leds/"
LED_PATHS = [
    "multicolor:chassis",
    ":rgb:joystick_rings",
]

logger = logging.getLogger(__name__)


def write_sysfs(dir: str, fn: str, val: Any):
    logger.info(f'Writing `{str(val)}` to \n"{os.path.join(dir, fn)}"')
    with open(os.path.join(dir, fn), "w") as f:
        f.write(str(val))


def read_sysfs(dir: str, fn: str, default: str | None = None):
    try:
        with open(os.path.join(dir, fn), "r") as f:
            return f.read().strip()
    except Exception as e:
        if default is not None:
            return default
        raise e


def get_led_path():
    for fn in os.listdir(LED_BASE):
        for p in LED_PATHS:
            if p in fn:
                return os.path.join(LED_BASE, fn)
    return None


def is_led_supported():
    return get_led_path() is not None


def chassis_led_set(ev: RgbLedEvent, init: bool = True):
    if ev["type"] != "led":
        return

    led_path = get_led_path()
    if not led_path:
        return

    match ev["mode"]:
        case "solid":
            r_mode = 1
        case _:
            r_mode = 0

    max_brightness = 255
    try:
        max_brightness = int(read_sysfs(led_path, "max_brightness", "255"))
    except Exception:
        pass

    r_brightness = max(min(int(ev["brightness"] * max_brightness), max_brightness), 0)
    r_red = max(min(ev["red"], 255), 0)
    r_green = max(min(ev["green"], 255), 0)
    r_blue = max(min(ev["blue"], 255), 0)

    # Mode only exists on ayn devices
    if init:
        try:
            write_sysfs(led_path, "led_mode", r_mode)
        except Exception:
            logger.info(
                "Could not write led_mode (not applicable for Ayaneo, only Ayn)."
            )
            try:
                write_sysfs(led_path, "device/led_mode", r_mode)
            except Exception:
                logger.info("Could not write led_mode to secondary path.")

        write_sysfs(led_path, "brightness", r_brightness)

    pattern = read_sysfs(led_path, "multi_index", "red green blue")
    arr = []
    for color in pattern.split(" "):
        if color == "red":
            arr.append(r_red)
        elif color == "green":
            arr.append(r_green)
        elif color == "blue":
            arr.append(r_blue)
        else:
            arr.append(0)

    write_sysfs(led_path, "multi_intensity", " ".join(map(str, arr)))


def thread_chassis_led_set(ev: RgbLedEvent, pending: TEvent, error: TEvent):
    try:
        chassis_led_set(ev)
    except Exception as e:
        logger.error(f"Setting leds failed with error:\n{e}")
        # Turn off support
        error.set()
    chassis_led_set(ev)
    pending.clear()


class LedDevice(Consumer):
    def __init__(self, rate_limit: float = 10, threading: bool = False) -> None:
        self.supported = is_led_supported()
        self.min_delay = 1 / rate_limit
        self.queued = None
        self.last = time.time() - self.min_delay

        self.threading = threading
        self.pending = TEvent()
        self.error = TEvent()
        self.t = None
        self.init = False

    def consume(self, events: Sequence[Event]):
        if not self.supported:
            return

        if self.error.isSet():
            self.supported = False
            return

        curr = time.time()
        ev = None

        # Pop queued event if possible
        if self.queued:
            e, t = self.queued
            if curr > t:
                e = ev
            self.queued = None

        # Find newer event if it exists
        for e in events:
            if e["type"] == "led":
                ev = e
                # Clear queue since there
                # is a newer event
                self.queued = None

        # If no led event return
        if ev is None:
            return

        if curr > self.last + self.min_delay and not self.pending.is_set():
            if self.threading:
                self.pending.set()
                self.t = Thread(
                    target=thread_chassis_led_set, args=(ev, self.pending, self.error)
                )
                self.t.start()
            else:
                try:
                    chassis_led_set(ev, not self.init)
                    self.init = True
                except Exception as e:
                    logger.error(f"Setting leds failed with error:\n{e}")
                    # Turn off support
                    self.supported = False
            self.last = curr
        else:
            self.queued = (ev, curr + self.min_delay)



import logging
from enum import Enum
from typing import Literal, Sequence

from hhd.controller import Event
from hhd.controller.lib.hid import Device
from hhd.controller.physical.hidraw import GenericGamepadHidraw

logger = logging.getLogger(__name__)

Controller = Literal["left", "right"]
RgbMode = Literal["solid", "pulse", "dynamic", "spiral"]

RGB_MODE_PULSE = 0x02
RGB_MODE_DYNAMIC = 0x03
RGB_MODE_SPIRAL = 0x04


def to_bytes(s: str):
    return bytes.fromhex(s.replace(" ", ""))


def _get_controller(c: Controller):
    if c == "left":
        return 0x03
    elif c == "right":
        return 0x04
    assert False, f"Controller '{c}' not supported."


def rgb_set_profile(
    controller: Controller,
    profile: Literal[1, 2, 3],
    mode: RgbMode,
    red: int,
    green: int,
    blue: int,
    brightness: float = 1,
    speed: float = 1,
):
    r_controller = _get_controller(controller)
    assert profile in (1, 2, 3), f"Invalid profile '{profile}' selected."

    match mode:
        case "solid":
            r_mode = 1
        case "pulse":
            r_mode = 2
        case "dynamic":
            r_mode = 3
        case "spiral":
            r_mode = 4
        case _:
            assert False, f"Mode '{mode}' not supported. "

    r_brightness = min(max(int(64 * brightness), 0), 63)
    r_period = min(max(int(64 * (1 - speed)), 0), 63)

    return bytes(
        [
            0x05,
            0x0C,
            0x72,
            0x01,
            r_controller,
            r_mode,
            red,
            green,
            blue,
            r_brightness,
            r_period,
            profile,
            0x01,
        ]
    )


def rgb_load_profile(
    controller: Controller,
    profile: Literal[1, 2, 3],
):
    r_controller = _get_controller(controller)

    return bytes(
        [
            0x05,
            0x06,
            0x73,
            0x02,
            r_controller,
            profile,
            0x01,
        ]
    )


def rgb_enable(controller: Controller, enable: bool):
    r_enable = enable & 0x01
    r_controller = _get_controller(controller)
    return bytes(
        [
            0x05,
            0x06,
            0x70,
            0x02,
            r_controller,
            r_enable,
            0x01,
        ]
    )


def controller_enable_gyro(controller: Controller):
    rc = _get_controller(controller)
    EN = 0x01
    M = 0x02
    return [
        # Enable the gyro if its disabled
        bytes([0x05, 0x06, 0x6A, 0x02, rc, EN, 0x01]),
        # Enable high quality report
        bytes([0x05, 0x06, 0x6A, 0x07, rc, M, 0x01]),
    ]


def controller_disable_gyro(controller: Controller):
    rc = _get_controller(controller)
    M = 0x01
    return [
        # Disable high quality report
        bytes([0x05, 0x06, 0x6A, 0x07, rc, M, 0x01]),
    ]


def controller_factory_reset():
    return [
        # RX
        to_bytes("0405 05 01 01 01 01"),
        # Dongle (?)
        to_bytes("0405 05 01 01 02 01"),
        # Left
        to_bytes("0405 05 01 01 03 01"),
        # Right
        to_bytes("0405 05 01 01 04 01"),
    ]


def controller_legion_swap(enabled):
    return [to_bytes(f"0506 69 0401 {'02' if enabled else '01'} 01")]


def rgb_multi_load_settings(
    mode: RgbMode,
    profile: Literal[1, 2, 3],
    red: int,
    green: int,
    blue: int,
    brightness: float = 1,
    speed: float = 1,
    init: bool = True,
):
    base = [
        rgb_set_profile("left", profile, mode, red, green, blue, brightness, speed),
        rgb_set_profile("right", profile, mode, red, green, blue, brightness, speed),
    ]
    # Always update
    # Old firmware has issues with new way
    # if not init:
    #     return base

    return [
        *base,
        rgb_load_profile("left", profile),
        rgb_load_profile("right", profile),
        rgb_enable("left", True),
        rgb_enable("right", True),
    ]


def rgb_multi_disable():
    return [
        rgb_enable("left", False),
        rgb_enable("right", False),
    ]


class RgbCallback:
    def __init__(self) -> None:
        self.prev_mode = None

    def __call__(self, dev: Device, events: Sequence[Event]):
        try:
            for ev in events:
                if ev["type"] != "led":
                    continue

                reps = None
                mode = None
                match ev["mode"]:
                    case "disabled":
                        pass
                    case "pulse":
                        mode = "pulse"
                    case "rainbow":
                        mode = "dynamic"
                    case "solid":
                        if ev["red"] or ev["green"] or ev["blue"]:
                            mode = "solid"
                        else:
                            # Disable if brightness is 0
                            mode = None
                    case "spiral":
                        mode = "spiral"
                    case _:
                        pass

                if mode:
                    reps = rgb_multi_load_settings(
                        mode,
                        0x03,
                        ev["red"],
                        ev["green"],
                        ev["blue"],
                        ev["brightness"],
                        ev["speed"],
                        self.prev_mode != mode,
                    )
                    # Only init sparingly, to speed up execution
                    self.prev_mode = mode
                else:
                    reps = rgb_multi_disable()

                for r in reps:
                    dev.write(r)
        except Exception as e:
            logger.error(f"Error while setting RGB:\n{e}")


class LegionHidraw(GenericGamepadHidraw):
    def with_settings(
        self,
        gyro: str | None,
        reset: bool,
        use_touchpad: bool = False,
        swap_legion: bool = False,
    ):
        self.gyro = gyro
        self.reset = reset
        self.use_touchpad = use_touchpad
        self.old_touched = False
        self.old_x = 0
        self.old_y = 0
        self.touch_changed = None
        self.touchpad_init = True
        self.swap_legion = swap_legion
        return self

    def open(self):
        out = super().open()
        if not out:
            return out
        if not self.dev:
            return out

        cmds = []

        if self.reset:
            logger.warning(f"Factory Resetting controllers")
            cmds.extend(controller_factory_reset())
        gyro_active = self.gyro in ("left", "right", "both")
        if self.gyro in ("left", "both") or (not gyro_active and self.use_touchpad):
            cmds.extend(controller_enable_gyro("left"))
        if self.gyro in ("right", "both"):
            cmds.extend(controller_enable_gyro("right"))
        # Always disable legion swap and use our version
        cmds.extend(controller_legion_swap(False))

        for r in cmds:
            self.dev.write(r)

        return out

    def produce(self, fds: Sequence[int]):
        out = super().produce(fds)

        # TODO: Cleanup
        # Or remove, since this option is problematic
        # If removing, remove the produce function completely
        if self.use_touchpad and self.report:
            if self.touchpad_init:
                self.touchpad_init = False
                out = [
                    *out,
                    {
                        "type": "configuration",
                        "code": "touchpad_aspect_ratio",
                        "value": 1,
                    },
                ]

            x = int.from_bytes(self.report[26:28])
            y = int.from_bytes(self.report[28:30])
            touched = bool(x or y)
            changed = touched != self.old_touched or x != self.old_x or y != self.old_y
            if changed:
                out = list(out)
                if touched != self.old_touched:
                    self.old_touched = touched
                    out.append(
                        {"type": "button", "code": "touchpad_touch", "value": touched}
                    )
                if x != self.old_x:
                    self.old_x = x
                    out.append(
                        {"type": "axis", "code": "touchpad_x", "value": x / 1000}
                    )
                if y != self.old_y:
                    self.old_y = y
                    out.append(
                        {"type": "axis", "code": "touchpad_y", "value": y / 1000}
                    )

        return out

    def close(self, exit: bool) -> bool:
        if not self.dev:
            return super().close(exit)

        cmds = []
        # Always reset both gyros to avoid leaving them on
        # in case they use battery
        cmds.extend(controller_disable_gyro("left"))
        cmds.extend(controller_disable_gyro("right"))
        # Restore the windows legion swap version for continuity
        cmds.extend(controller_legion_swap(self.swap_legion))
        for r in cmds:
            self.dev.write(r)

        return super().close(exit)



from hhd.controller.physical.evdev import B

MSI_CLAW_MAPPINGS = {
    B("KEY_VOLUMEUP"): "key_volumeup",
    B("KEY_VOLUMEDOWN"): "key_volumedown",
    B("KEY_F15"): "mode",
    B("KEY_F16"): "share",
}

CONFS = { 
    "MS-1T41": {
        "name": "MSI Claw (1st gen)",
        # "hrtimer": True, Uses sensor fusion garbage? From intel? Will need custom work
        "extra_buttons": "dual",
        "btn_mapping": MSI_CLAW_MAPPINGS,
    },
    "MS-1T42": {
        "name": "MSI Claw 7 (2nd gen)",
        "extra_buttons": "dual",
        "btn_mapping": MSI_CLAW_MAPPINGS,
    },
    "MS-1T52": {
        "name": "MSI Claw 8",
        "extra_buttons": "dual",
        "btn_mapping": MSI_CLAW_MAPPINGS,
    },
    "MS-1T8K": {
        "name": "MSI Claw A8",
        "extra_buttons": "dual",
        "btn_mapping": MSI_CLAW_MAPPINGS,
    },
}



from hhd.controller.lib.common import AM, BM

SDCONT_VENDOR = 0x28DE
SDCONT_VERSION = 256
SDCONT_COUNTRY = 0

SDCONT_DESCRIPTOR = bytes(
    [
        0x06,
        0xFF,
        0xFF,  # // Usage Page (Vendor Usage Page 0xffff) 0
        0x09,
        0x01,  # // Usage (Vendor Usage 0x01)           3
        0xA1,
        0x01,  # // Collection (Application)            5
        0x09,
        0x02,  # //  Usage (Vendor Usage 0x02)          7
        0x09,
        0x03,  # //  Usage (Vendor Usage 0x03)          9
        0x15,
        0x00,  # //  Logical Minimum (0)                11
        0x26,
        0xFF,
        0x00,  # //  Logical Maximum (255)              13
        0x75,
        0x08,  # //  Report Size (8)                    16
        0x95,
        0x40,  # //  Report Count (64)                  18
        0x81,
        0x02,  # //  Input (Data,Var,Abs)               20
        0x09,
        0x06,  # //  Usage (Vendor Usage 0x06)          22
        0x09,
        0x07,  # //  Usage (Vendor Usage 0x07)          24
        0x15,
        0x00,  # //  Logical Minimum (0)                26
        0x26,
        0xFF,
        0x00,  # //  Logical Maximum (255)              28
        0x75,
        0x08,  # //  Report Size (8)                    31
        0x95,
        0x40,  # //  Report Count (64)                  33
        0xB1,
        0x02,  # //  Feature (Data,Var,Abs)             35
        0xC0,  # // End Collection                      37
    ]
)

SD_AXIS_MAP = {
    "touchpad_x": AM((20 << 3), "i16", scale=2**16-3, offset=-2**15+1),
    "touchpad_y": AM((22 << 3), "i16", scale=-2**16+3, offset=2**15-1),
    "touchpad_force": AM((58 << 3), "i16", scale=2**14 - 2),
    "accel_x": AM(
        (24 << 3), "i16", scale=16384 / 9.80665, bounds=(-(2**15) + 2, 2**15 - 1)
    ),
    "accel_y": AM(
        (26 << 3), "i16", scale=-16384 / 9.80665, bounds=(-(2**15) + 2, 2**15 - 1)
    ),
    "accel_z": AM(
        (28 << 3), "i16", scale=16384 / 9.80665, bounds=(-(2**15) + 2, 2**15 - 1)
    ),
    "gyro_x": AM((30 << 3), "i16", scale=16 * 180 / 3.14),
    "gyro_z": AM((32 << 3), "i16", scale=16 * 180 / 3.14),
    "gyro_y": AM((34 << 3), "i16", scale=16 * 180 / 3.14),
    "rt": AM((44 << 3), "i16"),
    "lt": AM((46 << 3), "i16"),
    "ls_x": AM((48 << 3), "i16"),
    "ls_y": AM((50 << 3), "i16", flipped=True),
    "rs_x": AM((52 << 3), "i16"),
    "rs_y": AM((54 << 3), "i16", flipped=True),
}

# 0 -> 10000000 (0x80)
# 1 -> 01000000 (0x40)
# 2 -> 00100000 (0x20)
# 3 -> 00010000 (0x10)
# 4 -> 00001000 (0x08)
# 5 -> 00000100 (0x04)
# 6 -> 00000010 (0x02)
# 7 -> 00000001 (0x01)

SD_BTN_MAP = {
    "a": BM((8 << 3)),
    "x": BM((8 << 3) + 1),
    "b": BM((8 << 3) + 2),
    "y": BM((8 << 3) + 3),
    "lb": BM((8 << 3) + 4),
    "rb": BM((8 << 3) + 5),
    "rt": BM((8 << 3) + 6),
    "lt": BM((8 << 3) + 7),
    "extra_l2": BM((9 << 3)),
    "start": BM((9 << 3) + 1),
    "mode": BM((9 << 3) + 2),
    "select": BM((9 << 3) + 3),
    "dpad_down": BM((9 << 3) + 4),
    "dpad_left": BM((9 << 3) + 5),
    "dpad_right": BM((9 << 3) + 6),
    "dpad_up": BM((9 << 3) + 7),
    "ls": BM((10 << 3) + 1),
    "touchpad_touch": BM((10 << 3) + 3),
    "touchpad_left": BM((10 << 3) + 5),
    "extra_r2": BM((10 << 3) + 7),
    "rs": BM((11 << 3) + 5),
    "extra_r1": BM((13 << 3) + 5),
    "extra_l1": BM((13 << 3) + 6),
    "share": BM((14 << 3) + 5),
    # "touchpad_touch": BM((32 << 3), flipped=True),
    # "touchpad_touch2": BM((36 << 3), flipped=True),
    # "touchpad_left": BM((9 << 3) + 6),
}

SD_SETTINGS = [
    # /* 0 */
    "SETTING_MOUSE_SENSITIVITY",
    "SETTING_MOUSE_ACCELERATION",
    "SETTING_TRACKBALL_ROTATION_ANGLE",
    "SETTING_HAPTIC_INTENSITY_UNUSED",
    "SETTING_LEFT_GAMEPAD_STICK_ENABLED",
    "SETTING_RIGHT_GAMEPAD_STICK_ENABLED",
    "SETTING_USB_DEBUG_MODE",
    "SETTING_LEFT_TRACKPAD_MODE",
    "SETTING_RIGHT_TRACKPAD_MODE",
    "SETTING_MOUSE_POINTER_ENABLED",
    # /* 10 */
    "SETTING_DPAD_DEADZONE",
    "SETTING_MINIMUM_MOMENTUM_VEL",
    "SETTING_MOMENTUM_DECAY_AMMOUNT",
    "SETTING_TRACKPAD_RELATIVE_MODE_TICKS_PER_PIXEL",
    "SETTING_HAPTIC_INCREMENT",
    "SETTING_DPAD_ANGLE_SIN",
    "SETTING_DPAD_ANGLE_COS",
    "SETTING_MOMENTUM_VERTICAL_DIVISOR",
    "SETTING_MOMENTUM_MAXIMUM_VELOCITY",
    "SETTING_TRACKPAD_Z_ON",
    # /* 20 */
    "SETTING_TRACKPAD_Z_OFF",
    "SETTING_SENSITIVY_SCALE_AMMOUNT",
    "SETTING_LEFT_TRACKPAD_SECONDARY_MODE",
    "SETTING_RIGHT_TRACKPAD_SECONDARY_MODE",
    "SETTING_SMOOTH_ABSOLUTE_MOUSE",
    "SETTING_STEAMBUTTON_POWEROFF_TIME",
    "SETTING_UNUSED_1",
    "SETTING_TRACKPAD_OUTER_RADIUS",
    "SETTING_TRACKPAD_Z_ON_LEFT",
    "SETTING_TRACKPAD_Z_OFF_LEFT",
    # /* 30 */
    "SETTING_TRACKPAD_OUTER_SPIN_VEL",
    "SETTING_TRACKPAD_OUTER_SPIN_RADIUS",
    "SETTING_TRACKPAD_OUTER_SPIN_HORIZONTAL_ONLY",
    "SETTING_TRACKPAD_RELATIVE_MODE_DEADZONE",
    "SETTING_TRACKPAD_RELATIVE_MODE_MAX_VEL",
    "SETTING_TRACKPAD_RELATIVE_MODE_INVERT_Y",
    "SETTING_TRACKPAD_DOUBLE_TAP_BEEP_ENABLED",
    "SETTING_TRACKPAD_DOUBLE_TAP_BEEP_PERIOD",
    "SETTING_TRACKPAD_DOUBLE_TAP_BEEP_COUNT",
    "SETTING_TRACKPAD_OUTER_RADIUS_RELEASE_ON_TRANSITION",
    # /* 40 */
    "SETTING_RADIAL_MODE_ANGLE",
    "SETTING_HAPTIC_INTENSITY_MOUSE_MODE",
    "SETTING_LEFT_DPAD_REQUIRES_CLICK",
    "SETTING_RIGHT_DPAD_REQUIRES_CLICK",
    "SETTING_LED_BASELINE_BRIGHTNESS",
    "SETTING_LED_USER_BRIGHTNESS",
    "SETTING_ENABLE_RAW_JOYSTICK",
    "SETTING_ENABLE_FAST_SCAN",
    "SETTING_IMU_MODE",
    "SETTING_WIRELESS_PACKET_VERSION",
    # /* 50 */
    "SETTING_SLEEP_INACTIVITY_TIMEOUT",
    "SETTING_TRACKPAD_NOISE_THRESHOLD",
    "SETTING_LEFT_TRACKPAD_CLICK_PRESSURE",
    "SETTING_RIGHT_TRACKPAD_CLICK_PRESSURE",
    "SETTING_LEFT_BUMPER_CLICK_PRESSURE",
    "SETTING_RIGHT_BUMPER_CLICK_PRESSURE",
    "SETTING_LEFT_GRIP_CLICK_PRESSURE",
    "SETTING_RIGHT_GRIP_CLICK_PRESSURE",
    "SETTING_LEFT_GRIP2_CLICK_PRESSURE",
    "SETTING_RIGHT_GRIP2_CLICK_PRESSURE",
    # /* 60 */
    "SETTING_PRESSURE_MODE",
    "SETTING_CONTROLLER_TEST_MODE",
    "SETTING_TRIGGER_MODE",
    "SETTING_TRACKPAD_Z_THRESHOLD",
    "SETTING_FRAME_RATE",
    "SETTING_TRACKPAD_FILT_CTRL",
    "SETTING_TRACKPAD_CLIP",
    "SETTING_DEBUG_OUTPUT_SELECT",
    "SETTING_TRIGGER_THRESHOLD_PERCENT",
    "SETTING_TRACKPAD_FREQUENCY_HOPPING",
    # /* 70 */
    "SETTING_HAPTICS_ENABLED",
    "SETTING_STEAM_WATCHDOG_ENABLE",
    "SETTING_TIMP_TOUCH_THRESHOLD_ON",
    "SETTING_TIMP_TOUCH_THRESHOLD_OFF",
    "SETTING_FREQ_HOPPING",
    "SETTING_TEST_CONTROL",
    "SETTING_HAPTIC_MASTER_GAIN_DB",
    "SETTING_THUMB_TOUCH_THRESH",
    "SETTING_DEVICE_POWER_STATUS",
    "SETTING_HAPTIC_INTENSITY",
    # /* 80 */
    "SETTING_STABILIZER_ENABLED",
    "SETTING_TIMP_MODE_MTE",
]



def evdev(dev: str | None):
    from typing import cast
    from evdev import list_devices, InputDevice, categorize, ecodes
    from time import sleep, perf_counter

    cache = {}

    def B(b: str):
        if b not in cache:
            cache[b] = getattr(ecodes, b)
        return cast(int, cache[b])

    def RV(type: int, code: int):
        if (type, code) not in cache:
            try:
                v = ecodes.bytype[type][code]
            except KeyError:
                v = "UNKWN"
            if isinstance(v, list) or isinstance(v, tuple):
                v = v[0]
            cache[(type, code)] = v
        else:
            v = cache[(type, code)]
        return v

    def EVT(type):
        if type not in cache:
            v = getattr(ecodes, "EV")[type]
            cache[type] = v
        else:
            v = cache[type]
        return v

    print("Available Devices with the Current Permissions")
    avail = list_devices()
    for d in avail:
        print(f" - {str(InputDevice(d))}")
    print()

    if dev and dev != "nograb":
        print(f"Using argument '{dev}'.")
        try:
            sel = f"/dev/input/event{int(dev)}"
        except Exception:
            sel = dev
        if sel not in avail:
            print(f"Device '{sel}' not found.")
            return
    else:
        sel =    None
        while sel not in avail:
            try:
                sel = input("Enter device path (/dev/input/event# or #): ")
            except EOFError:
                return
            try:
                sel = f"/dev/input/event{int(sel)}"
            except Exception as e:
                pass
        print()

    d = InputDevice(sel)
    print(f"Selected device `{d}`.")
    print()
    print("Capabilities")
    for (cap_str, cap), vals in d.capabilities(verbose=True).items():
        print(f" - {cap_str} ({cap:x})")
        for (names, code) in vals:
            if not isinstance(code, int):
                abs_info = code
                names, code = names
            else:
                abs_info = ""
            print(f"   0x{code:04x}: {', '.join(names) if isinstance(names, list) else names}")
            if abs_info:
                print(f"     > [{str(abs_info)}]")
    try:
        print()
        if dev != "nograb":
            print("Attempting to grab device.")
            d.grab()
            print("Device grabbed, system will not see its events.")
    except Exception as e:
        print(f"Could not grab device, system will still see events. Error:\n{e}")
        print("\nReading events still work.")
    print()
    print(f"Reading from: `{d}`.")

    endcap = True
    start = perf_counter()
    ofs = None
    prev = 0
    out = ""
    for ev in d.read_loop():
        if ofs == None:
            ofs = ev.timestamp()
        curr = perf_counter() - start
        hz = f"{1/(curr - prev):6.1f} Hz" if prev and curr != prev else "   NaN Hz"
        if ev.code == 0 and ev.type == 0 and ev.value == 0:
            out += (
                f"└ SYN ─ {curr:7.3f}s ─ {hz} ───────────────────────────────────────┘"
            )
            print(out)
            out = ""
            prev = curr
            endcap = True
        else:
            if endcap:
                out += "\n┌───────────────────────────────────────────────────────────────────┐\n"
                endcap = False

            evstr = (
                f"{ev.timestamp() - ofs:7.3f}s /"
                + f" {EVT(ev.type):>6s} ({ev.type:02x}) /"
                + f" {RV(ev.type, ev.code):>21s} (x{ev.code:03x}):"
            )

            if ev.type == B("EV_KEY"):
                match ev.value:
                    case 0:
                        act = "released"
                    case 1:
                        act = " pressed"
                    case 2:
                        act = "repeated"
                    case val:
                        act = f"{val:8d}"
                evstr += f" {act}"
            elif ev.type == B("EV_ABS"):
                evstr += f" {ev.value:8d}"
            else:
                hexval = f"{ev.value:04X}"
                evstr += f" {hexval:>8s}"

            out += f"│ {evstr:>65s} │\n"
        sleep(0.00005)

def device_str(d):
    from hhd.controller.lib.common import hexify

    return (
        f"{d['path'].decode():13s} {hexify(d['vendor_id'])}:{hexify(d['product_id'])}"
        + f" Usage Page: 0x{hexify(d['usage_page'])} Usage: 0x{hexify(d['usage'])}"
        + f" Interface: {d['interface_number']} Names: '{d['manufacturer_string']}': '{d['product_string']}'"
    )

def hidraw(dev: str | None, *cmds: str):
    from hhd.controller.lib.hid import enumerate_unique, Device
    from time import sleep, perf_counter

    avail = []
    infos = {}
    devs = {}
    for i, d in enumerate(enumerate_unique()):
        avail.append(d["path"])
        n = int(d["path"].decode().split("hidraw")[1])
        infos[n] = (
            f" - {device_str(d)}"
        )
        devs[d['path']] = d

    if dev:
        print(f"Using argument '{dev}'.")
        try:
            sel = f"/dev/hidraw{int(dev)}".encode()
        except Exception:
            sel = dev.encode()
        if sel not in avail or not sel:
            print(f"Device '{sel.decode()}' not found.")
            return
    else:
        print("Available Devices with the Current Permissions")
        print("\n".join([infos[k] for k in sorted(infos)]))
        print()

        sel = None
        while not sel or sel not in avail:
            try:
                sel = input("Enter device path (/dev/hidraw# or #): ")
            except EOFError:
                return
            try:
                sel = f"/dev/hidraw{int(sel)}".encode()
            except Exception:
                sel = sel.encode()
        print()

    d = Device(path=sel)

    if cmds:
        print(f"Device: {device_str(devs[sel])}")
        print()
        print(f"Writing provided commands to device:")
        for cmd in cmds:
            # Cleanup and get type
            cmd = cmd.lower().strip()
            if cmd.startswith('set:'):
                cmd_type = "set"
                cmd = cmd[4:]
            elif cmd.startswith('get:'):
                cmd_type = "get"
                cmd = cmd[4:]
            else:
                cmd_type = "write"
            cmd = cmd.replace(' ', '').replace(':', '')

            match cmd_type:
                case "write":
                    print(f" - {cmd}")
                    try:
                        d.write(bytes.fromhex(cmd))
                    except Exception as e:
                        print(f"Error writing command '{cmd}':\n{e}")
                        return
                case "set":
                    print(f" - SET {cmd}")
                    try:
                        d.send_feature_report(bytes.fromhex(cmd))
                    except Exception as e:
                        print(f"Error setting feature '{cmd}':\n{e}")
                        return
                case "get":
                    print(f" - GET {cmd}")
                    try:
                        print(d.get_feature_report(int(cmd, 16)).hex())
                    except Exception as e:
                        print(f"Error getting feature '{cmd}':\n{e}")
                        return
        return

    print(f"Device Information:")
    for k, v in devs[sel].items():
        print(f" - {k:>20s}: {f'0x{v:04x}' if isinstance(v, int) else v}")

    try:
        from .hid_desc import print_descriptor
        print('\nDevice HID Descriptor:')
        print_descriptor(d.fd)
    except Exception as e:
        print(f"Could not get descriptor:\n{e}")

    print()
    print(f"Selected device:\n{device_str(devs[sel])}\n")

    start = perf_counter()
    prev = 0
    for i in range(100000000):
        curr = perf_counter() - start
        hz = f"{1/(curr - prev):6.1f} Hz" if prev and curr != prev else "   NaN Hz"
        prev = curr
        print(f"{i:6d}: {curr:8.4f}s ({hz})", d.read().hex())
        sleep(0.0005)



import argparse
import json
import logging
import socket
import sys
from http.client import HTTPConnection

logger = logging.getLogger(__name__)

SOCKET_UNIX = "/run/hhd/api"
USAGE = """
hhdctl [-h] [--sep SEP] [--values] {get,set,poll,track} [keys ...]

Handheld Daemon CLI
This CLI is used to interact with  Handheld Daemon (hhd) via its API. It requires 
root access to connect to the UNIX socket at /run/hhd/api. The current version 
allows for querying the state of Handheld Daemon and updating its values.

Commands:
    get: Get the current state of Handheld Daemon. If keys are provided, only 
        those keys are returned. Returned as KEY=VALUE pairs separated by \\n.
    set: Update values in the current state. This call blocks until values are
        updated and returns the new values. WARNING: the new values might not
        be the ones that were set if they were rejected. Use None to remove a key.
    poll: Same as get but will wait for the next Handheld Daemon event loop to
        return. The loop runs every 2s or whenever an event is received.
    track: Continuously track the provided values. Same as calling get and then 
        poll repeatedly. The separator between updates can be changed with --sep. 
        Default is \\n. track will print the values even if they did not change.

Examples:
    hhdctl get
    hhdctl get/track/poll rgb.handheld.mode.mode
    hhdctl set rgb.handheld.mode.mode=oxp
    # For a single value, --values and --sep='' can be used to return the value
    hhdctl get rgb.handheld.mode.mode --values --sep=''
"""


def _unroll_dict(d, prefix=""):
    if isinstance(d, dict):
        for k, v in d.items():
            if prefix:
                k = f"{prefix}.{k}"
            yield from _unroll_dict(v, k)
    else:
        yield prefix, d


def unroll_dict(d):
    return dict(_unroll_dict(d))


class UnixConnection(HTTPConnection):
    def connect(self):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(SOCKET_UNIX)


def _request(*args, **kwargs):
    con = UnixConnection(SOCKET_UNIX)
    con.request(*args, **kwargs)
    return con.getresponse()


def _get_state(poll: bool = False):
    return _request("GET", f"/api/v1/state{'?poll' if poll else ''}")


def _set_state(state):
    return _request("POST", f"/api/v1/state", body=json.dumps(state))


def _get(keys, poll: bool = False, state=None, values: bool = False):
    if state is not None:
        state = _set_state(state)
    else:
        state = _get_state(poll)

    # Drop values in keys for ease of use
    if keys:
        keys = [k.split("=", 1)[0] for k in keys]

    if state.status != 200:
        logger.error(f"Failed to get state with status: {state.status}")
        return 2

    out = ""
    data = unroll_dict(json.loads(state.read()))
    err = 0
    for k in keys or data:
        if k not in data or data[k] is None:
            # Eat none values as the purge happens after restart
            logger.error(f"Key {k} not found in state")
            err = 3
            continue
        else:
            vr = data[k]

        if isinstance(vr, bool):
            v = "true" if vr else "false"
        elif isinstance(vr, str):
            # Escape newlines
            v = vr.replace("\n", "\\n").replace("\t", "\\t")
        else:
            v = vr

        if values:
            out += f"{v}\n"
        else:
            out += f"{k}={v}\n"

    sys.stdout.write(out)
    sys.stdout.flush()
    return err


def _track(keys, sep, values):
    poll = False
    while True:
        _get(keys, poll=poll, values=values)
        sys.stdout.write(sep)
        sys.stdout.flush()
        poll = True


def _set(keys, values):
    if not keys:
        logger.error("No keys provided to set")
        return 1

    state = {}
    for key in keys:
        try:
            k, vr = key.split("=", 1)
        except ValueError:
            logger.error(f"Invalid KEY=VALUE: {key}")
            return 1

        if vr.lower() in ["true", "false"]:
            v = vr.lower() == "true"
        elif vr.lower() == "none":
            v = None
        elif vr.isdigit():
            v = int(vr)
        elif vr.isnumeric():
            v = float(vr)
        else:
            v = vr

        state[k] = v

    return _get(keys, state=state, values=values)


def _main():
    logging.basicConfig(
        level=logging.DEBUG, stream=sys.stderr, format="%(levelname)s - %(message)s"
    )

    parser = argparse.ArgumentParser(
        usage=USAGE,
        prog="hhdctl",
    )

    parser.add_argument(
        "command", help="Command to execute", choices=["get", "set", "poll", "track"]
    )
    parser.add_argument(
        "keys",
        help="Key(s) to get/set/track. Format: KEY=VAL for set and KEY for get, track. If not provided, get and track return all parameters.",
        nargs="*",
    )
    parser.add_argument(
        "--sep",
        help="Separator for updates in track",
        default="\n",
    )
    parser.add_argument(
        "--values",
        help="Hide the KEY= part in the response. The return value is 3 if a value is missing.",
        action="store_true",
        default=False,
    )

    args = parser.parse_args()

    match args.command:
        case "get":
            v = _get(args.keys, values=args.values)
        case "set":
            v = _set(args.keys, args.values)
        case "track":
            v = _track(args.keys, args.sep, args.values)
        case "poll":
            v = _get(args.keys, poll=True, values=args.values)
        case _:
            logger.error(f"Invalid command: '{args.command}'")
            v = -1

    if v is not None:
        sys.exit(v)


def set_state(state):
    res = _set_state(state)
    if res.status != 200:
        raise Exception(f"Failed to set state with status: {res.status}")
    return json.loads(res.read())


def get_state(poll: bool = False):
    res = _get_state(poll)
    if res.status != 200:
        raise Exception(f"Failed to get state with status: {res.status}")
    return json.loads(res.read())


def main():
    try:
        _main()
    except KeyboardInterrupt:
        sys.exit(0)
    except PermissionError:
        logger.error(
            "Permission denied when trying to connect to the UNIX socket. Are you running as root?"
        )
        sys.exit(1)


ALL = {
    "set_state": set_state,
    "get_state": get_state,
    "unroll_dict": unroll_dict,
    "main": main,
}


if __name__ == "__main__":
    main()



import argparse
import sys
import time
from .ctl import get_state, set_state, unroll_dict

SOCKET_UNIX = "/run/hhd/api"
USAGE = """
hhd.steamos [-h] {steamos-select-branch,steamos-update} [--fallback] [keys ...]

Handheld Daemon steamos polkit stub
Allows mimicking the polkit behavior of SteamOS to perform updates, etc.
For specifics, refer to SteamOS. The --fallback option is provided which will
return 20 if handheld daemon cannot update the system. In this case, you can
use the legacy fallback to update the system.

Commands:
    steamos-select-branch: Select a branch that running steamos-update will update to.
        Aliased to steamos-branch-select. Options are: rel, rc, beta, main, bc, -l, -c.
    steamos-update: Perform an update. 

"""


ALL = {
    "set_state": set_state,
    "get_state": get_state,
}

BRANCH_MAP = {
    "rel": "stable",
    "beta": "testing",
    "preview": "testing",
    "rc": "testing",
    "bc": "unstable",
    "pc": "unstable",
    "main": "unstable",
}

FALLBACK_CODE = 20


def _select_branch(fallback, opts):
    branch = "stable"
    try:
        state = unroll_dict(get_state())
        stage = state.get("updates.bootc.steamos-update", None)
        # print(f"Stage: {stage}", file=sys.stderr)
        incompatible = stage is None or stage == "incompatible"

        img = state.get("updates.bootc.image", None).split(":")[-1]
        assert img is not None
        for k, v in BRANCH_MAP.items():
            if img.startswith(v):
                branch = k
                break
    except Exception as e:
        incompatible = True
        print(f"Error: {e}", file=sys.stderr)
    if incompatible and fallback:
        return FALLBACK_CODE

    if "-c" in opts:
        print(branch)
        return 0
    if "-l" in opts:
        for v in BRANCH_MAP.keys():
            print(v)
        return 0

    if not opts:
        print("No option provided", file=sys.stderr)
        return 0

    if incompatible:
        print("Incompatible state", file=sys.stderr)
        return 0

    if not opts:
        return 0

    target = opts[0]
    if target == branch:
        return 0

    target_os = None
    for k, v in BRANCH_MAP.items():
        if target == k:
            target_os = v
            break

    if target_os is None:
        print(f"Invalid branch: {target}", file=sys.stderr)
        return 0
    
    print("Ignoring request to rebase from SteamOS", file=sys.stderr)
    return 0


def _update(fallback, opts):
    if "--supports-duplicate-detection" in opts:
        return 0

    check = "check" in opts

    # Check if there is an update
    try:
        if unroll_dict(get_state()).get("updates.bootc.steamos-update", None) in (
            "incompatible",
            None,
        ):
            print("Incompatible state", file=sys.stderr)
            return FALLBACK_CODE if fallback else 0

        val = unroll_dict(set_state({"updates.bootc.steamos-update": "check"})).get(
            "updates.bootc.steamos-update", None
        )
        while val == "check":
            val = unroll_dict(get_state(poll=True)).get(
                "updates.bootc.steamos-update", None
            )
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        if fallback:
            return FALLBACK_CODE
        return 1

    if val != "has-update" and not (val and val.endswith("%")):
        print(f"No updates available ({val})", file=sys.stderr)
        return 7
    elif check:
        print(f"Updates available ({val})", file=sys.stderr)
        return 0

    # Otherwise apply, bootc does not need separate steps
    if not val or not val.endswith("%"):
        val = unroll_dict(set_state({"updates.bootc.steamos-update": "apply"})).get(
            "updates.bootc.steamos-update", None
        )

    curr = 0.2
    while not val or "%" in val or val == "apply":
        if val and val.endswith("%"):
            next = float(val[:-1])
            while curr < next:
                # print(f"\r\033[K\r{curr:.2f}% 1m1s", end="")
                print(f"\r{curr:.2f}%", end="")
                sys.stdout.flush()
                curr += min(5, next - curr)
                time.sleep(0.2)

        val = unroll_dict(get_state(poll=True)).get(
            "updates.bootc.steamos-update", None
        )

    if val != "updated":
        return 1
    return 0


def main():
    fallback = False
    try:
        if "--help" in sys.argv or "-h" in sys.argv or len(sys.argv) < 2:
            print(USAGE)
            sys.exit(0)

        fallback = "--fallback" in sys.argv
        cmd = sys.argv[1]

        opts = [v for v in sys.argv[2:] if v != "--fallback"]
        match cmd:
            case "steamos-branch-select" | "steamos-select-branch":
                v = _select_branch(fallback, opts)
            case "steamos-update":
                v = _update(fallback, opts)
            case _:
                print(f"Invalid command: '{cmd}'")
                v = -1

        if v is not None:
            sys.exit(v)

    except KeyboardInterrupt:
        sys.exit(0)
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



from hhd.controller import Axis, Button, Configuration
from hhd.controller.physical.evdev import B, to_map
from hhd.plugins import gen_gyro_state

DEFAULT_MAPPINGS: dict[str, tuple[Axis, str | None, float, float | None]] = {
    "accel_x": ("accel_x", "accel", 1, None),
    "accel_y": ("accel_z", "accel", 1, None),
    "accel_z": ("accel_y", "accel", -1, None),
    "anglvel_x": ("gyro_x", "anglvel", 1, None),
    "anglvel_y": ("gyro_z", "anglvel", 1, None), 
    "anglvel_z": ("gyro_y", "anglvel", -1, None),
    "timestamp": ("imu_ts", None, 1, None),
}

AT_BTN_MAPPINGS: dict[int, str] = {
    # Volume buttons come from the same keyboard
    B("KEY_F16"): "mode",  # Big Button
    B("KEY_F15"): "share",  # Small Button
    # B("KEY_F17"): "extra_l1",  # LC Button
    # B("KEY_F18"): "extra_r1",  # RC Button
}

GAMEPAD_BTN_MAPPINGS: dict[int, str] = {
    # Volume buttons come from the same keyboard
    # B("KEY_F16"): "mode",  # Big Button
    # B("KEY_F15"): "share",  # Small Button
    B("KEY_F17"): "extra_l1",  # LC Button
    B("KEY_F18"): "extra_r1",  # RC Button
}

CONFS = {
    # New hardware new firmware, the unit below was dissassembled
    # "G1621-02": {"name": "OrangePi G1621-02/G1621-02", "hrtimer": True},
    "NEO-01": {"name": "OrangePi NEO-01/NEO-01", "hrtimer": True},
}


def get_default_config(product_name: str):
    out = {
        "name": product_name,
        "hrtimer": True,
        "untested": True,
    }

    return out



from copy import deepcopy
from threading import Lock
from typing import (
    Any,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
    TypeVar,
    cast,
)

Pytree = int | float | str | Sequence["Pytree"] | Mapping[str, "Pytree"]
A = TypeVar("A")


def parse_conf(c: Pytree, out: MutableMapping | None = None):
    if not isinstance(c, MutableMapping):
        return c

    if out is None:
        out = {}

    for multival, v in c.items():
        d = out
        subs = multival.split(".")
        for k in subs[:-1]:
            d[k] = d.get(k, {})
            d = d[k]

        if subs[-1] not in d or (
            not isinstance(d[subs[-1]], Mapping) or not isinstance(v, Mapping)
        ):
            d[subs[-1]] = parse_conf(v)
        else:
            mout = {}
            parse_conf(d[subs[-1]], mout)
            parse_conf(v, mout)
            d[subs[-1]] = mout

    return out


def parse_confs(
    confs: Sequence[Pytree], out: MutableMapping | Pytree | None = None
) -> MutableMapping | Pytree:
    if not isinstance(out, MutableMapping):
        out = {}
    for c in confs:
        if not isinstance(out, MutableMapping):
            out = {}
        if isinstance(c, Mapping):
            parse_conf(c, out)
        else:
            out = c
    return out


def to_seq(key: str | tuple[str, ...]):
    if isinstance(key, str):
        key = (key,)

    seq = []
    for k in key:
        for s in k.split("."):
            seq.append(s)
    return seq


def compare_dicts(a, b):
    if len(a) != len(b):
        return False

    for k in a:
        if not k in b:
            return False

        if isinstance(a[k], Mapping) and isinstance(b[k], Mapping):
            if not compare_dicts(a[k], b[k]):
                return False
        else:
            if not a[k] == b[k]:
                return False

    return True


class Config:
    def __init__(
        self, conf: Pytree | Sequence[Pytree] = [], readonly: bool = False
    ) -> None:
        self._conf: Pytree | MutableMapping = {}
        self._lock = Lock()
        self._updated = False
        self.readonly = readonly
        self.update(conf)
        self.updated = False

    def update(self, conf: Pytree | Sequence[Pytree]):
        with self._lock:
            conf = deepcopy(conf)
            if isinstance(conf, Sequence):
                self._conf = parse_confs(conf, self._conf)
            else:
                if isinstance(self._conf, MutableMapping):
                    parse_conf(conf, self._conf)
                else:
                    self._conf = conf
        self.updated = True

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, Config):
            return False

        if __value is self:
            return True

        with __value._lock, self._lock:
            return compare_dicts(__value._conf, self._conf)

    def __setitem__(self, key: str | tuple[str, ...], val):
        with self._lock:
            val = deepcopy(val)
            seq = to_seq(key)

            cont = {}
            d = cont
            for s in seq[:-1]:
                d[s] = {}
                d = d[s]

            d[seq[-1]] = val
            if isinstance(self._conf, MutableMapping):
                init = deepcopy(self._conf)
                parse_conf(cont, self._conf)
                # FIXME: verify no regressions
                if init != self._conf:
                    self._updated = True
            else:
                self._conf = cont
                if self._conf != cont:
                    self._updated = True

    def __contains__(self, key: str | tuple[str, ...]):
        with self._lock:
            seq = to_seq(key)
            d = self._conf
            for s in seq:
                if s not in d:
                    return False
                d = cast(Mapping, d)[s]
        return True

    def __getitem__(self, key: str | tuple[str, ...]) -> "Config":
        with self._lock:
            assert isinstance(self._conf, MutableMapping)
            seq = to_seq(key)
            d = self._conf
            for s in seq:
                d = cast(Mapping, d)[s]
            return Config([deepcopy(d)])

    def __delitem__(self, key: str | tuple[str, ...]):
        with self._lock:
            assert isinstance(self._conf, MutableMapping)
            seq = to_seq(key)
            d = self._conf
            for s in seq[:-1]:
                d = cast(Mapping, d)[s]
            del d[seq[-1]]
        self.updated = True

    def get(self, key, default: A) -> A:
        try:
            return self[key].to(type(default))
        except KeyError:
            return default
        except TypeError:
            return default

    def get_action(self, key):
        if key not in self:
            return False
        tmp = bool(self[key].conf)
        if tmp:
            self[key] = False
        return tmp

    def to(self, t: type[A]) -> A:
        return cast(t, self.conf)

    def copy(self):
        return Config([self.conf])

    @property
    def conf(self):
        with self._lock:
            return deepcopy(self._conf)

    @property
    def updated(self):
        with self._lock:
            return self._updated

    @updated.setter
    def updated(self, v: bool):
        with self._lock:
            self._updated = v



import logging
import select
import time
from threading import Event as TEvent

from hhd.controller import DEBUG_MODE, Multiplexer, can_read
from hhd.controller.lib.hide import unhide_all
from hhd.controller.physical.hidraw import GenericGamepadHidraw
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import GenericGamepadEvdev, enumerate_evs
from hhd.controller.virtual.uinput import UInputDevice
from hhd.plugins import Config, Context, Emitter, get_outputs
from hhd.controller.physical.evdev import DINPUT_AXIS_POSTPROCESS, AbsAxis
from hhd.controller.physical.evdev import (
    GamepadButton,
    GenericGamepadEvdev,
    enumerate_evs,
    to_map,
)
from typing import Sequence
from hhd.controller import DEBUG_MODE, Event, Multiplexer

from .const import MSI_CLAW_MAPPINGS

FIND_DELAY = 0.1
ERROR_DELAY = 0.3
LONGER_ERROR_DELAY = 3
LONGER_ERROR_MARGIN = 1.3

logger = logging.getLogger(__name__)

CLAW_SET_M1M2 = lambda a, btn: bytes(
    [0x0F, 0x00, 0x00, 0x3C, 0x21, 0x01, *a[btn], 0x05, 0x01, 0x00, 0x00, 0x12, 0x00]
)
CLAW_SYNC_ROM = bytes([0x0F, 0x00, 0x00, 0x3C, 0x22])

CLAW_SET_DINPUT = bytes([0x0F, 0x00, 0x00, 0x3C, 0x24, 0x02, 0x00])
CLAW_SET_MSI = bytes([0x0F, 0x00, 0x00, 0x3C, 0x24, 0x03, 0x00])
CLAW_SET_DESKTOP = bytes([0x0F, 0x00, 0x00, 0x3C, 0x24, 0x04, 0x00])

MSI_CLAW_VID = 0x0DB0
MSI_CLAW_XINPUT_PID = 0x1901
MSI_CLAW_DINPUT_PID = 0x1902
MSI_CLAW_TEST_PID = 0x1903

KBD_VID = 0x0001
KBD_PID = 0x0001

BACK_BUTTON_DELAY = 0.1

# 0211
ADDR_0163 = {
    "rgb": [0x01, 0xFA],
    "m1": [0x00, 0x7A],
    "m2": [0x01, 0x1F],
}
# 0217
# 0308 (?)
ADDR_0166 = {
    "rgb": [0x02, 0x4A],
    "m1": [0x00, 0xBA],
    "m2": [0x01, 0x63],
}
ADDR_DEFAULT = ADDR_0166


def set_rgb_cmd(brightness, red, green, blue, addr: dict = ADDR_DEFAULT) -> bytes:
    return bytes(
        [
            # Preamble
            0x0F,
            0x00,
            0x00,
            0x3C,
            # Write first profile
            0x21,
            0x01,
            # Start at
            *addr["rgb"],
            # Write 31 bytes
            0x20,
            # Index, Frame num, Effect, Speed, Brightness
            0x00,
            0x01,
            0x09,
            0x03,
            max(0, min(100, int(brightness * 100))),
        ]
    ) + 9 * bytes([red, green, blue])


class ClawDInputHidraw(GenericGamepadHidraw):

    def __init__(self, *args, test_mode: bool = False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.init = False
        self.test_mode = test_mode
        self.addr = None

    def open(self):
        out = super().open()
        if not out:
            return out

        if self.addr is None:
            ver = (self.info or {}).get("release_number", 0x0)
            major = ver >> 8
            logger.info(f"Device version: {ver:#04x}")
            if (
                (major == 1 and ver >= 0x0166)
                or (major == 2 and ver >= 0x0217)
                or (major >= 3)
            ):
                self.addr = ADDR_0166
            else:
                self.addr = ADDR_0163

        return out

    def write(self, cmd: bytes) -> None:
        if not self.dev:
            return

        self.dev.write(cmd + bytes([0x00] * ((64 - len(cmd)) if len(cmd) < 64 else 0)))
        logger.info(f"Sent command: {cmd.hex()}")
        # # Receive ack
        # for _ in range(10):
        #     # 10 00 00 3c 06
        #     cmd = self.dev.read(64)
        #     if cmd and cmd[4] == 0x06:
        #         break

    def consume(self, events: Sequence[Event]) -> None:
        if not self.dev:
            return

        for ev in events:
            if ev["type"] == "rumble" and ev["code"] == "main":
                # Same as dualshock 4
                # "0501 0000 right left"
                cmd = bytes(
                    [
                        0x05,
                        0x01,
                        0x00,
                        0x00,
                        min(255, int(ev["weak_magnitude"] * 255)),
                        min(255, int(ev["strong_magnitude"] * 255)),
                        00,
                        00,
                        00,
                        00,
                        00,
                    ]
                )
                self.dev.write(cmd)
            elif ev["type"] == "led" and not self.test_mode:
                if ev["mode"] == "solid":
                    cmd = set_rgb_cmd(
                        ev["brightness"],
                        ev["red"],
                        ev["green"],
                        ev["blue"],
                        self.addr or ADDR_DEFAULT,
                    )
                    self.write(cmd)
                elif ev["mode"] == "disabled":
                    cmd = set_rgb_cmd(
                        0,
                        0,
                        0,
                        0,
                        self.addr or ADDR_DEFAULT,
                    )
                    self.write(cmd)

    def set_dinput_mode(self, init: bool = False) -> None:
        if not self.dev:
            return

        # Make sure M1/M2 are recognizable
        if init:
            time.sleep(0.3)
            self.write(CLAW_SET_M1M2(self.addr or ADDR_DEFAULT, "m1"))
            time.sleep(0.5)
            self.write(CLAW_SET_M1M2(self.addr or ADDR_DEFAULT, "m2"))
            time.sleep(0.5)
            self.write(CLAW_SYNC_ROM)
            time.sleep(0.5)
            self.write(CLAW_SET_MSI)
            time.sleep(2)

        # Set the device to dinput mode
        self.write(CLAW_SET_DINPUT)


DINPUT_BUTTON_MAP: dict[int, GamepadButton] = to_map(
    {
        # Gamepad
        "a": [EC("BTN_B")],
        "b": [EC("BTN_C")],
        "x": [EC("BTN_A")],
        "y": [EC("BTN_NORTH")],
        # Sticks
        "ls": [EC("BTN_SELECT")],
        "rs": [EC("BTN_START")],
        # Bumpers
        "lb": [EC("BTN_WEST")],
        "rb": [EC("BTN_Z")],
        # Select
        "start": [EC("BTN_TR2")],
        "select": [EC("BTN_TL2")],
        # Misc
        "extra_l1": [EC("BTN_TRIGGER_HAPPY")],
        "extra_r1": [0x013F],
    }
)
DINPUT_AXIS_MAP: dict[int, AbsAxis] = to_map(
    {
        # Sticks
        # Values should range from -1 to 1
        "ls_x": [EC("ABS_X")],
        "ls_y": [EC("ABS_Y")],
        "rs_x": [EC("ABS_Z")],
        "rs_y": [EC("ABS_RZ")],
        # Triggers
        # Values should range from -1 to 1
        "rt": [EC("ABS_RX")],
        "lt": [EC("ABS_RY")],
        # Hat, implemented as axis. Either -1, 0, or 1
        "hat_x": [EC("ABS_HAT0X")],
        "hat_y": [EC("ABS_HAT0Y")],
    }
)


def plugin_run(
    conf: Config,
    emit: Emitter,
    context: Context,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
    woke_up: TEvent,
):
    first = True
    first_disabled = True
    initialized = False
    init = time.perf_counter()
    repeated_fail = False
    while not should_exit.is_set():
        if conf["controller_mode.mode"].to(str) == "disabled":
            time.sleep(ERROR_DELAY)
            if first_disabled:
                UInputDevice.close_volume_cached()
                unhide_all()
            first_disabled = False
            continue
        else:
            first_disabled = True

        try:
            is_xinput = bool(enumerate_evs(vid=MSI_CLAW_VID, pid=MSI_CLAW_XINPUT_PID))
            found_device = bool(
                enumerate_evs(vid=MSI_CLAW_VID, pid=MSI_CLAW_DINPUT_PID)
            )
            test_mode = bool(enumerate_evs(vid=MSI_CLAW_VID, pid=MSI_CLAW_TEST_PID))
        except Exception:
            logger.warning("Failed finding device, skipping check.")
            time.sleep(LONGER_ERROR_DELAY)
            found_device = True
            is_xinput = False

        if is_xinput or (found_device and not test_mode and not initialized):
            d_vend = ClawDInputHidraw(
                vid=[MSI_CLAW_VID],
                pid=[MSI_CLAW_XINPUT_PID, MSI_CLAW_DINPUT_PID],
                application=[
                    0xFFF00040,
                    0x00010005,
                    0xFF000000,
                    0xFFA00001,
                ],
                required=True,
            )
            initialized = True
            try:
                d_vend.open()
                d_vend.set_dinput_mode(init=True)
                d_vend.close(True)
                time.sleep(2)
            except Exception as e:
                logger.error(f"Failed to set device into dinput mode.\n{type(e)}: {e}")
                time.sleep(1)

        if not found_device and not test_mode:
            if first:
                logger.info("Controller not found. Waiting...")
            time.sleep(FIND_DELAY)
            first = False
            continue

        try:
            logger.info("Launching emulated controller.")
            updated.clear()
            init = time.perf_counter()
            controller_loop(
                conf.copy(),
                should_exit,
                updated,
                dconf,
                emit,
                woke_up,
                test_mode=test_mode,
            )
            repeated_fail = False
        except Exception as e:
            failed_fast = init + LONGER_ERROR_MARGIN > time.perf_counter()
            sleep_time = (
                LONGER_ERROR_DELAY if repeated_fail and failed_fast else ERROR_DELAY
            )
            repeated_fail = failed_fast
            logger.error(f"Received the following error:\n{type(e)}: {e}")
            logger.error(
                f"Assuming controllers disconnected, restarting after {sleep_time}s."
            )
            first = True
            # Raise exception
            if DEBUG_MODE:
                raise e
            time.sleep(sleep_time)

    # Unhide all devices before exiting and close keyboard cache
    UInputDevice.close_volume_cached()
    unhide_all()


class DesktopDetectorEvdev(GenericGamepadEvdev):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.desktop = False

    def produce(self, fds: Sequence[int]):
        if not self.dev or self.fd not in fds:
            return []

        while can_read(self.fd):
            for e in self.dev.read():
                self.desktop = True

        return []


def controller_loop(
    conf: Config,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
    emit: Emitter,
    woke_up: TEvent,
    test_mode: bool = False,
):
    debug = DEBUG_MODE

    logger.info(f"Test mode: {test_mode}")
    # Output
    d_producers, d_outs, d_params = get_outputs(
        conf["controller_mode"],
        None,
        emit=emit,
        rgb_modes={"disabled": [], "solid": ["color"]},
    )

    # Inputs
    d_xinput = GenericGamepadEvdev(
        vid=[MSI_CLAW_VID],
        pid=[MSI_CLAW_DINPUT_PID, MSI_CLAW_TEST_PID],
        # name=["Generic X-Box pad"],
        capabilities={EC("EV_KEY"): [EC("BTN_A")]},
        required=True,
        hide=True,
        btn_map=DINPUT_BUTTON_MAP,
        axis_map=DINPUT_AXIS_MAP,
        postprocess=DINPUT_AXIS_POSTPROCESS,
    )

    d_kbd_1 = GenericGamepadEvdev(
        vid=[KBD_VID],
        pid=[KBD_PID],
        required=False,
        grab=True,
        btn_map=dconf.get("btn_mapping", MSI_CLAW_MAPPINGS),
    )

    # Mute these so after suspend we do not get stray keypresses
    d_kbd_2 = DesktopDetectorEvdev(
        vid=[MSI_CLAW_VID],
        pid=[MSI_CLAW_DINPUT_PID],
        required=False,
        grab=True,
        capabilities={EC("EV_KEY"): [EC("KEY_ESC")]},
    )
    d_mouse = DesktopDetectorEvdev(
        vid=[MSI_CLAW_VID],
        pid=[MSI_CLAW_DINPUT_PID],
        required=False,
        grab=True,
        capabilities={EC("EV_KEY"): [EC("BTN_MOUSE")]},
    )

    kargs = {}

    multiplexer = Multiplexer(
        trigger="analog_to_discrete",
        dpad="analog_to_discrete",
        share_to_qam=True,
        select_reboots=conf.get("select_reboots", False),
        nintendo_mode=conf["nintendo_mode"].to(bool),
        emit=emit,
        params=d_params,
        startselect_chord=conf.get("main_chords", "disabled"),
        swap_guide="select_is_guide" if conf["swap_guide"].to(bool) else None,
        keyboard_no_release=True,
        **kargs,
    )

    d_volume_btn = UInputDevice(
        name="Handheld Daemon Volume Keyboard",
        phys="phys-hhd-vbtn",
        capabilities={EC("EV_KEY"): [EC("KEY_VOLUMEUP"), EC("KEY_VOLUMEDOWN")]},
        btn_map={
            "key_volumeup": EC("KEY_VOLUMEUP"),
            "key_volumedown": EC("KEY_VOLUMEDOWN"),
        },
        pid=KBD_PID,
        vid=KBD_VID,
        output_timestamps=True,
    )

    if test_mode:
        d_vend = ClawDInputHidraw(
            vid=[MSI_CLAW_VID],
            pid=[MSI_CLAW_TEST_PID],
            application=[
                0xFFF00040,
                0x00010005,
                0xFF000000,
            ],
            required=True,
            test_mode=True,
        )
    else:
        d_vend = ClawDInputHidraw(
            vid=[MSI_CLAW_VID],
            pid=[MSI_CLAW_DINPUT_PID],
            application=[
                0xFFF00040,
                0x00010005,
                0xFF000000,
            ],
            required=True,
        )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 400

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        prepare(d_xinput)
        prepare(d_volume_btn)
        prepare(d_kbd_1)
        if not test_mode:
            prepare(d_kbd_2)
            prepare(d_mouse)
        for d in d_producers:
            prepare(d)
        prepare(d_vend)

        logger.info("Emulated controller launched, have fun!")
        switch_to_dinput = None
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))

            # Detect if we are in desktop mode through events
            desktop_mode = d_mouse.desktop or d_kbd_2.desktop
            d_mouse.desktop = False
            d_kbd_2.desktop = False

            if desktop_mode or (switch_to_dinput and start > switch_to_dinput):
                logger.info("Setting controller to dinput mode.")
                d_vend.set_dinput_mode()
                switch_to_dinput = None
            # elif woke_up.is_set():
            #     woke_up.clear()
            #     # Switch to dinput after 2 seconds without input to avoid
            #     # being stuck in desktop mode
            #     switch_to_dinput = time.perf_counter() + 2

            evs = multiplexer.process(evs)
            if evs:
                switch_to_dinput = None
                if debug:
                    logger.info(evs)

                d_volume_btn.consume(evs)
                d_xinput.consume(evs)
                d_vend.consume(evs)

            for d in d_outs:
                d.consume(evs)

            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        d_vend.close(not updated.is_set())
        for d in reversed(devs):
            try:
                d.close(not updated.is_set())
            except Exception as e:
                logger.error(f"Error while closing device '{d}' with exception:\n{e}")
                if debug:
                    raise e



import logging
import subprocess
from typing import Literal
import os

logger = logging.getLogger(__name__)


class WakeHandler:
    def __init__(self) -> None:
        self.proc = None
        self.inhibitor = None
        self.broken = False
        self.fd = -1
        self.got_prepare = False

    def start(self):
        try:
            self.proc = subprocess.Popen(
                [
                    "dbus-monitor",
                    "--system",
                    "type='signal',interface='org.freedesktop.login1.Manager'",
                    "--monitor",
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
                text=True,
            )
            assert self.proc.stdout is not None
            self.fd = self.proc.stdout.fileno()
            os.set_blocking(self.fd, False)

            if not self.inhibit(True):
                # We need both the inhibitor and reader for this to work
                self.close()
                return False

            return True
        except Exception:
            if self.proc:
                self.proc.terminate()
                self.proc.wait()
            self.proc = None
            self.broken = True
            return False

    def inhibit(self, enable: bool):
        if self.inhibitor:
            self.inhibitor.terminate()
            self.inhibitor.wait()
            self.inhibitor = None

        if self.broken:
            return False

        if enable:
            try:
                self.inhibitor = subprocess.Popen(
                    [
                        "systemd-inhibit",
                        "--what=sleep",
                        "--mode=delay",
                        "--who",
                        "HandheldDaemon",
                        "--why",
                        "Handheld Daemon: Turn off display",
                        "--",
                        "tail",
                        "-f",
                        "/dev/null",
                    ],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
                return True
            except Exception:
                self.inhibitor = None
                return False

        return True

    def __call__(self) -> Literal["entry", "exit", None]:
        if self.broken or self.fd == -1 or not self.proc or not self.proc.stdout:
            return None
        try:
            while line := self.proc.stdout.readline():
                if "PrepareForSleep" in line:
                    self.got_prepare = True
                elif self.got_prepare and "boolean" in line:
                    self.got_prepare = False
                    return "entry" if "true" in line else "exit"
        except Exception as e:
            logger.error(f"Systemd monitor error:\n{e}")
            self.close()
            self.broken = True

        return None

    def close(self):
        if self.proc:
            self.proc.kill()
            self.proc.wait()
        self.proc = None
        self.inhibit(False)



import logging
from typing import TYPE_CHECKING, Any, Sequence

from hhd.plugins import Config, Context, HHDPlugin, load_relative_yaml

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .const import PowerButtonConfig

from threading import Event, Thread


def run(**config: Any):
    from .base import power_button_run

    power_button_run(**config)


class PowerbuttondPlugin(HHDPlugin):
    def __init__(self, cfg: "PowerButtonConfig") -> None:
        self.name = f"powerbuttond@'{cfg.device}'"
        self.priority = 90
        self.log = "pbtn"
        self.cfg = cfg
        self.t = None
        self.event = None
        self.emit = None

    def open(
        self,
        emit,
        context: Context,
    ):
        self.started = False
        self.context = context
        self.emit = emit

    def settings(self):
        d = {"hhd": load_relative_yaml("settings.yml")}
        # if self.cfg.unsupported:
        #     d["hhd"]["settings"]["children"]["powerbuttond"]["default"] = False
        return d

    def update(self, conf: Config):
        if conf["hhd.settings.powerbuttond"].to(bool) and not self.started:
            self.start()
        elif not conf["hhd.settings.powerbuttond"].to(bool) and self.started:
            self.stop()
            logger.info('Stopping Steam Powerbutton Handler.')

    def start(self):
        from .base import power_button_run

        self.event = Event()
        self.t = Thread(
            target=power_button_run, args=(self.cfg, self.context, self.event, self.emit)
        )
        self.t.start()
        self.started = True

    def stop(self):
        if not self.event or not self.t:
            return
        self.event.set()
        self.t.join()
        self.event = None
        self.t = None
        self.started = False

    def close(self):
        self.stop()


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    from .const import get_config

    return [PowerbuttondPlugin(get_config())]



import logging
import time
from collections import deque
from typing import Literal

from hhd.controller import can_read
from hhd.controller.physical.hidraw import GenericGamepadHidraw

logger = logging.getLogger(__name__)


def gen_cmd(cid: int, cmd: bytes | list[int] | str, size: int = 64):
    # Command: [idx, cid, 0x3f, *cmd, 0x3f, cid], idx is optional
    if isinstance(cmd, str):
        c = bytes.fromhex(cmd)
    else:
        c = bytes(cmd)
    base = bytes([cid, 0xFF, *c])
    return base + bytes([0] * (size - len(base)))


def gen_rgb_mode(mode: str):
    mc = 0
    match mode:
        case "monster_woke":
            mc = 0x0D
        case "flowing":
            mc = 0x03
        case "sunset":
            mc = 0x0B
        case "neon":
            mc = 0x05
        case "dreamy":
            mc = 0x07
        case "cyberpunk":
            mc = 0x09
        case "colorful":
            mc = 0x0C
        case "aurora":
            mc = 0x01
        case "sun":
            mc = 0x08
    return gen_cmd(0x07, [mc])


gen_intercept = lambda enable: gen_cmd(0xB2, [0x03 if enable else 0x00, 0x01, 0x02])


def gen_brightness(
    enabled: bool,
    brightness: Literal["low", "medium", "high"],
):
    match brightness:
        case "low":
            bc = 0x01
        case "medium":
            bc = 0x03
        case _:  # "high":
            bc = 0x04

    return gen_cmd(0x07, [0xFD, enabled, 0x05, bc])


def gen_rgb_solid(r, g, b):
    return gen_cmd(0x07, [0xFE] + 20 * [r, g, b] + [0x00])


KBD_NAME = "keyboard"
HOME_NAME = "guide"
KBD_NAME_NON_TURBO = "share"
KBD_HOLD = 0.3
OXP_BUTTONS = {
    0x24: KBD_NAME,
    0x21: HOME_NAME,
    0x22: "extra_l1",
    0x23: "extra_r1",
}


INITIALIZE = [
    # gen_cmd(
    #     0xF5,
    #     "010238020101010101000000020102000000030103000000040104000000050105000000060106000000070107000000080108000000090109000000",
    # ),
    # gen_cmd(
    #     0xF5,
    #     "0102380202010a010a0000000b010b0000000c010c0000000d010d0000000e010e0000000f010f000000100110000000220200000000230200000000",
    # ),
    # gen_intercept(False),
]

INIT_DELAY = 4
WRITE_DELAY = 0.05
SCAN_DELAY = 1


class OxpHidrawV2(GenericGamepadHidraw):
    def __init__(self, *args, turbo: bool = True, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.prev = {}
        self.queue_kbd = None
        self.queue_home = None
        self.queue_cmd = deque(maxlen=10)
        self.next_send = 0
        self.queue_led = None
        self.turbo = turbo

        self.prev_brightness = None
        self.prev_stick = None
        self.prev_stick_enabled = None
        # self.prev_center = None
        # self.prev_center_enabled = None

    def open(self):
        a = super().open()
        self.queue_kbd = None
        self.queue_home = None
        self.prev = {}
        self.next_send = time.perf_counter() + INIT_DELAY

        self.queue_cmd.extend(INITIALIZE)
        return a

    def consume(self, events):
        if not self.dev:
            return

        # Capture led events
        for ev in events:
            if ev["type"] == "led":
                # if self.queue_led:
                #     logger.warning("OXP HID LED event queue overflow.")
                self.queue_led = ev

        # Send queued event if applicable
        curr = time.perf_counter()
        if self.queue_cmd and curr - self.next_send > 0:
            cmd = self.queue_cmd.popleft()
            logger.info(f"OXP C: {cmd.hex()}")
            self.dev.write(cmd)
            self.next_send = curr + WRITE_DELAY

        # Queue needs to flush before switching to next event
        # Also, there needs to be a led event to queue
        if self.queue_cmd or not self.queue_led:
            return
        ev = self.queue_led
        self.queue_led = None

        brightness = "high"
        stick = None
        stick_enabled = True

        match ev["mode"]:
            case "solid":
                stick = ev["red"], ev["green"], ev["blue"]
            case "oxp":
                brightness = ev["brightnessd"]
                stick = ev["oxp"]
                if stick == "classic":
                    # Classic mode is a cherry red
                    stick = 0xB7, 0x30, 0x00
            case _:  # "disabled":
                stick_enabled = False

        # Force RGB to not initialize to workaround RGB breaking
        # rumble when being set
        if self.prev_stick_enabled is None:
            self.prev_stick_enabled = stick_enabled
        if self.prev_brightness is None:
            self.prev_brightness = brightness
        if self.prev_stick is None:
            self.prev_stick = stick

        if (
            stick_enabled != self.prev_stick_enabled
            or brightness != self.prev_brightness
        ):
            self.queue_cmd.append(gen_brightness(stick_enabled, brightness))
            self.prev_brightness = brightness
            self.prev_stick_enabled = stick_enabled

        if stick_enabled and stick != self.prev_stick:
            if isinstance(stick, str):
                self.queue_cmd.append(gen_rgb_mode(stick))
            else:
                self.queue_cmd.append(gen_rgb_solid(*stick))
            self.prev_stick = stick
            self.prev_brightness = brightness
            self.prev_stick_enabled = stick_enabled

    def produce(self, fds):
        if not self.dev:
            return []

        evs = []
        # A bit unclean with 2 buttons but it works
        if self.queue_kbd:
            curr = time.perf_counter()
            if curr - KBD_HOLD > self.queue_kbd:
                evs = [
                    {
                        "type": "button",
                        "code": KBD_NAME if self.turbo else KBD_NAME_NON_TURBO,
                        "value": False,
                    }
                ]
                self.queue_kbd = None
        if self.queue_home:
            curr = time.perf_counter()
            if curr - KBD_HOLD > self.queue_home:
                evs = [
                    {
                        "type": "button",
                        "code": HOME_NAME,
                        "value": False,
                    }
                ]
                self.queue_home = None

        if self.fd not in fds:
            return evs

        while can_read(self.fd):
            cmd = self.dev.read()
            # logger.info(f"OXP R: {cmd.hex()}")

            cid = cmd[0]
            valid = cmd[1] == 0x3F and cmd[-2] == 0x3F

            if not valid:
                logger.warning(f"OXP HID invalid command: {cmd.hex()}")
                continue

            if cid in (0xF5, 0xB8):
                # Initialization (0xf5) and rgb (0xb8) command responses, skip
                continue

            if cid != 0xB2:
                logger.warning(f"OXP HID unknown command: {cmd.hex()}")
                continue

            btn = cmd[6]

            if btn not in OXP_BUTTONS:
                logger.warning(
                    f"OXP HID unknown button: {btn:x} from cmd:\n{cmd.hex()}"
                )
                continue

            btn = OXP_BUTTONS[btn]
            pressed = cmd[12] == 1

            if btn == KBD_NAME:
                if pressed and (btn not in self.prev or self.prev[btn] != pressed):
                    evs.append(
                        {
                            "type": "button",
                            "code": KBD_NAME if self.turbo else KBD_NAME_NON_TURBO,
                            "value": True,
                        }
                    )
                    self.queue_kbd = time.perf_counter()
                self.prev[btn] = pressed
                continue

            if btn == HOME_NAME:
                if pressed and (btn not in self.prev or self.prev[btn] != pressed):
                    evs.append(
                        {
                            "type": "button",
                            "code": HOME_NAME,
                            "value": True,
                        }
                    )
                    self.queue_home = time.perf_counter()
                self.prev[btn] = pressed
                continue

            if btn in self.prev and self.prev[btn] == pressed:
                # Debounce
                continue

            self.prev[btn] = pressed
            evs.append(
                {
                    "type": "button",
                    "code": btn,
                    "value": pressed,
                }
            )

        return evs



import logging
import os
import select
import time
from threading import Event as TEvent

from hhd.controller import DEBUG_MODE, Multiplexer
from hhd.controller.lib.hide import unhide_all
from hhd.controller.physical.hidraw import GenericGamepadHidraw
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import GenericGamepadEvdev, enumerate_evs
from hhd.controller.physical.imu import CombinedImu, HrtimerTrigger
from hhd.controller.physical.rgb import LedDevice, is_led_supported
from hhd.controller.virtual.uinput import UInputDevice
from hhd.plugins import Config, Context, Emitter, get_gyro_state, get_outputs

from .const import BTN_MAPPINGS, DEFAULT_MAPPINGS, TECNO_RAW_INTERFACE_BTN_MAP

FIND_DELAY = 0.1
ERROR_DELAY = 0.3
LONGER_ERROR_DELAY = 3
LONGER_ERROR_MARGIN = 1.3

logger = logging.getLogger(__name__)


GAMEPAD_VID = 0x045E
GAMEPAD_PID = 0x028E

TECNO_VID = 0x2993
TECNO_PID = 0x2001

ZOTAC_VID = 0x1EE9
ZOTAC_PID = 0x1590

KBD_VID = 0x0001
KBD_PID = 0x0001

BACK_BUTTON_DELAY = 0.1


def plugin_run(
    conf: Config,
    emit: Emitter,
    context: Context,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
):
    first = True
    first_disabled = True
    init = time.perf_counter()
    repeated_fail = False
    while not should_exit.is_set():
        if conf["controller_mode.mode"].to(str) == "disabled":
            time.sleep(ERROR_DELAY)
            if first_disabled:
                UInputDevice.close_volume_cached()
                unhide_all()
            first_disabled = False
            continue
        else:
            first_disabled = True

        try:
            match dconf.get("type", None):
                case "tecno":
                    vid = TECNO_VID
                case "zotac":
                    vid = ZOTAC_VID
                case _:
                    vid = GAMEPAD_VID
            found_device = bool(enumerate_evs(vid=vid))
        except Exception:
            logger.warning("Failed finding device, skipping check.")
            time.sleep(LONGER_ERROR_DELAY)
            found_device = True

        if not found_device:
            if first:
                logger.info("Controller not found. Waiting...")
            time.sleep(FIND_DELAY)
            first = False
            continue

        try:
            logger.info("Launching emulated controller.")
            updated.clear()
            init = time.perf_counter()
            controller_loop(conf.copy(), should_exit, updated, dconf, emit)
            repeated_fail = False
        except Exception as e:
            failed_fast = init + LONGER_ERROR_MARGIN > time.perf_counter()
            sleep_time = (
                LONGER_ERROR_DELAY if repeated_fail and failed_fast else ERROR_DELAY
            )
            repeated_fail = failed_fast
            logger.error(f"Received the following error:\n{type(e)}: {e}")
            logger.error(
                f"Assuming controllers disconnected, restarting after {sleep_time}s."
            )
            first = True
            # Raise exception
            if DEBUG_MODE:
                raise e
            time.sleep(sleep_time)

    # Unhide all devices before exiting and close keyboard cache
    UInputDevice.close_volume_cached()
    unhide_all()


def controller_loop(
    conf: Config, should_exit: TEvent, updated: TEvent, dconf: dict, emit: Emitter
):
    debug = DEBUG_MODE
    dtype = dconf.get("type", "generic")
    dgyro = dconf.get("display_gyro", True)

    # Output
    d_producers, d_outs, d_params = get_outputs(
        conf["controller_mode"],
        None,
        dgyro and conf["imu"].to(bool),
        emit=emit,
        rgb_modes={"disabled": [], "solid": ["color"]} if is_led_supported() else None,
        rgb_resets_on_ac=is_led_supported(),
    )
    motion = d_params.get("uses_motion", True)

    # Imu
    if dgyro:
        d_imu = CombinedImu(
            conf["imu_hz"].to(int),
            get_gyro_state(conf["imu_axis"], dconf.get("mapping", DEFAULT_MAPPINGS)),
        )
    else:
        d_imu = None
    d_timer = HrtimerTrigger(conf["imu_hz"].to(int), [HrtimerTrigger.IMU_NAMES])

    # Inputs
    d_xinput = GenericGamepadEvdev(
        vid=[GAMEPAD_VID, TECNO_VID, ZOTAC_VID],
        pid=[GAMEPAD_PID, TECNO_PID, ZOTAC_PID],
        # name=["Generic X-Box pad"],
        capabilities={EC("EV_KEY"): [EC("BTN_A")]},
        required=True,
        hide=True,
    )

    d_kbd_1 = GenericGamepadEvdev(
        vid=[KBD_VID],
        pid=[KBD_PID],
        required=False,
        grab=True,
        btn_map=dconf.get("btn_mapping", BTN_MAPPINGS),
    )
    d_kbd_2 = None

    kargs = {}
    if dtype == "tecno":
        kargs = {
            "keyboard_is": "steam_qam",
            "qam_hhd": True,
        }

    multiplexer = Multiplexer(
        trigger="analog_to_discrete",
        dpad="analog_to_discrete",
        share_to_qam=True,
        select_reboots=conf.get("select_reboots", False),
        nintendo_mode=conf["nintendo_mode"].to(bool),
        emit=emit,
        params=d_params,
        startselect_chord=conf.get("main_chords", "disabled"),
        swap_guide="select_is_guide" if conf["swap_guide"].to(bool) else None,
        **kargs,
    )

    d_volume_btn = UInputDevice(
        name="Handheld Daemon Volume Keyboard",
        phys="phys-hhd-vbtn",
        capabilities={EC("EV_KEY"): [EC("KEY_VOLUMEUP"), EC("KEY_VOLUMEDOWN")]},
        btn_map={
            "key_volumeup": EC("KEY_VOLUMEUP"),
            "key_volumedown": EC("KEY_VOLUMEDOWN"),
        },
        pid=KBD_PID,
        vid=KBD_VID,
        output_timestamps=True,
    )

    d_rgb = LedDevice()
    if d_rgb.supported:
        logger.info(f"RGB Support activated through kernel driver.")

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 400

    if motion:
        REPORT_FREQ_MAX = max(REPORT_FREQ_MAX, conf["imu_hz"].to(float))

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        if dtype == "tecno":
            d_kbd_2 = GenericGamepadHidraw(
                vid=[TECNO_VID],
                pid=[TECNO_PID],
                usage_page=[0xFFA0],
                usage=[0x0001],
                required=True,
                btn_map=TECNO_RAW_INTERFACE_BTN_MAP,
            )
        if dtype == "zotac":
            d_kbd_1 = GenericGamepadEvdev(
                vid=[ZOTAC_VID],
                pid=[ZOTAC_PID],
                required=True,
                btn_map=dconf.get("btn_mapping", BTN_MAPPINGS),
                capabilities={EC("EV_KEY"): [EC("KEY_F17"), EC("KEY_F18")]},
            )

        prepare(d_xinput)
        if motion and d_imu:
            start_imu = True
            if dconf.get("hrtimer", False):
                start_imu = d_timer.open()
            if start_imu:
                prepare(d_imu)
        prepare(d_volume_btn)
        prepare(d_kbd_1)
        if d_kbd_2:
            prepare(d_kbd_2)
        for d in d_producers:
            prepare(d)

        logger.info("Emulated controller launched, have fun!")
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

                d_volume_btn.consume(evs)
                d_xinput.consume(evs)

            d_rgb.consume(evs)
            for d in d_outs:
                d.consume(evs)

            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        # d_vend.close(not updated.is_set())
        try:
            d_timer.close()
        except Exception as e:
            logger.error(f"Error while closing device '{d}' with exception:\n{e}")
            if debug:
                raise e
        for d in reversed(devs):
            try:
                d.close(not updated.is_set())
            except Exception as e:
                logger.error(f"Error while closing device '{d}' with exception:\n{e}")
                if debug:
                    raise e



import os
import logging

logger = logging.getLogger(__name__)

EV_KEY = 0x01
BTN_MISC = 0x100
BTN_JOYSTICK = 0x120
# KEY_MAX = 0x2ff
# BTN_TRIGGER_HAPPY = 0x2c0
BTN_GAMEPAD = 0x130
EV_ABS = 0x03
ABS_HAT0X = 0x10

BTN_FROM_SDL = {
    # Thumbpad
    "a": "a",
    "b": "b",
    "x": "x",
    "y": "y",
    # D-PAD
    "dpup": "dpad_up",
    "dpdown": "dpad_down",
    "dpleft": "dpad_left",
    "dpright": "dpad_right",
    # Sticks
    "leftstick": "ls",
    "rightstick": "rs",
    # Bumpers
    "leftshoulder": "lb",
    "rightshoulder": "rb",
    # Triggers
    "lefttrigger": "lt",
    "righttrigger": "rt",
    # Select
    "start": "start",
    "back": "select",
    # Misc
    "guide": "mode",
    "misc1": "share",
    # Back buttons
    "paddle1": "extra_l1",
    "paddle2": "extra_r1",
    "paddle3": "extra_l2",
    "paddle4": "extra_r2",
}

AXES_FROM_SDL = {
    "leftx": "ls_x",
    "lefty": "ls_y",
    "rightx": "rs_x",
    "righty": "rs_y",
    "lefttrigger": "lt",
    "righttrigger": "rt",
}

WAKE_KEYS = [
    "select",
    "mode",
    "b",
    "y",
    "a",
    "x",
]

OVERLAY_KEYS = [
    "a",
    "b",
    "x",
    "y",
    "lb",
    "rb",
    "mode",
    "select",
    "dpad_up",
    "dpad_down",
    "dpad_left",
    "dpad_right",
]

OVERLAY_AXES = [
    "ls_x",
    "ls_y",
    "rs_x",
    "rs_y",
]

CONTROLLERDB_FN = os.environ.get(
    "HHD_CONTROLLERDB", "/usr/share/sdl/gamecontrollerdb.txt"
)


def crc16_for_byte(byte):
    crc = 0
    for _ in range(8):
        if (byte ^ crc) & 1:
            crc = (crc >> 1) ^ 0xA001
        else:
            crc >>= 1
        byte >>= 1
    return crc


def sdl_crc16(crc, data):
    for byte in data:
        crc = crc16_for_byte(crc ^ byte) ^ (crc >> 8)
    return crc


def create_joystick_guid(v):
    # SDL_GUID SDL_CreateJoystickGUID function
    guid = bytearray(16)

    guid[0:2] = v["bus"].to_bytes(2, "little")
    guid[2:4] = sdl_crc16(0, v["name"].encode()).to_bytes(2, "little")
    if v["vendor"]:
        guid[4:6] = v["vendor"].to_bytes(2, "little")
        guid[8:10] = v["product"].to_bytes(2, "little")
        guid[12:14] = v["version"].to_bytes(2, "little")
        # Driver signatures, always 0
        # guid[14] = 0
        # guid[15] = 0
    else:
        # Here, there is a check for driver signature
        # However, sysjoystick (what we use) provides zeros
        guid[4:16] = v["name"].encode()[:12].ljust(12, b"\0")

    return guid


def get_joypad_buttons(dev):
    before_js = []
    after_js = []
    for i, b in enumerate(dev.get("byte", {}).get("key", bytes())):
        for j in range(8):
            if not (b & (1 << j)):
                continue
            key = i * 8 + j

            # Joydev kernel driver does the following\
            # but on e.g., 8bitdo KEY_MENU is used and missed
            # # Ignore smaller than BTN_MISC keys
            # if key < BTN_MISC:
            #     continue
            if key < BTN_JOYSTICK:
                before_js.append(key)
            else:
                after_js.append(key)

    return after_js + before_js


def get_joypad_axes(dev):
    abs = []
    hats = []
    for i, b in enumerate(dev.get("byte", {}).get("abs", bytes())):
        for j in range(8):
            if not (b & (1 << j)):
                continue
            key = i * 8 + j
            abs.append(key)
            if key >= ABS_HAT0X:
                hats.append(key)

    hats_x = [x for x in hats if x % 2 == 0]

    pairs = []
    for hat_x in hats_x:
        if hat_x + 1 in hats:
            pairs.append((hat_x, hat_x + 1))

    return abs, pairs


def load_mappings(fn: str = CONTROLLERDB_FN):
    mappings = {}

    try:
        with open(fn, "r") as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith("#"):
                    continue

                try:
                    guid_str, name, *bindings = line.split(",")
                    if guid_str == "xinput":
                        continue

                    # Check platform
                    different_platform = False
                    for bind in bindings:
                        if not bind.startswith("platform:"):
                            continue

                        platform = bind.split(":")[1].strip().lower()
                        if platform != "linux":
                            different_platform = True
                            break
                    if different_platform:
                        continue

                    mappings[bytes.fromhex(guid_str)] = (name, bindings)
                except Exception as e:
                    logger.info(f"Error parsing line '{line}': {e}")
    except Exception as e:
        logger.error(f"Failed to load SDL gamepad mappings from {fn}:\n{e}")
        return {}

    logger.info(f"Loaded {len(mappings)} SDL gamepad mappings from:\n{fn}")
    return mappings


def map_gamepad(bindings, jaxes, jbuttons, jhats):
    axes = {}
    buttons = {}

    for bind in bindings:
        if not bind:
            continue

        key, val = bind.split(":")

        if key == "crc" or key == "platform":
            continue

        hhd_btn = BTN_FROM_SDL.get(key, None)
        hhd_ax = AXES_FROM_SDL.get(key, None)
        if not (hhd_btn or hhd_ax):
            continue

        flip = False
        if val.startswith("-"):
            flip = True
            val = val[1:]
        if val.startswith("+"):
            val = val[1:]
        if val.endswith("~"):
            flip = True
            val = val[:-1]

        if val.startswith("a"):
            ax = jaxes[int(val[1:])]
            if hhd_ax:
                axes[ax] = {
                    "type": "axis",
                    "code": hhd_ax,
                    "flip": flip,
                }
            else:
                assert hhd_btn
                axes[ax] = {
                    "type": "button",
                    "code": hhd_btn,
                    "flip": flip,
                }

        elif val.startswith("b"):
            try:
                btn = jbuttons[int(val[1:])]
            except IndexError:
                # TODO: add error
                continue
            assert hhd_btn
            buttons[btn] = hhd_btn

        elif val.startswith("h"):
            hat_id, hat_ofs = val[1:].split(".")

            hat_x, hat_y = jhats[int(hat_id)]
            assert hhd_btn

            match int(hat_ofs):
                case 1:
                    axes[hat_y] = {
                        "type": "hat",
                        "up_code": axes.get(hat_y, {}).get("up_code", None),
                        "down_code": hhd_btn,
                    }
                case 2:
                    axes[hat_x] = {
                        "type": "hat",
                        "up_code": hhd_btn,
                        "down_code": axes.get(hat_x, {}).get("down_code", None),
                    }
                case 4:
                    axes[hat_y] = {
                        "type": "hat",
                        "up_code": hhd_btn,
                        "down_code": axes.get(hat_y, {}).get("down_code", None),
                    }
                case 8:
                    axes[hat_x] = {
                        "type": "hat",
                        "up_code": axes.get(hat_x, {}).get("up_code", None),
                        "down_code": hhd_btn,
                    }
                case _:
                    assert False, f"Unknown hat offset {hat_ofs}"

    return axes, buttons


def match_gamepad(device, mappings):
    guid = create_joystick_guid(device)
    # Matching CRC is not relevant for us
    # crc = int.from_bytes(guid[2:4], "little")
    guid[2:4] = b"\0\0"  # Clear CRC for matching

    match = mappings.get(bytes(guid), None)
    if not match:
        guid[12:14] = b"\0\0"  # Clear version for matching
        match = mappings.get(bytes(guid), None)

    if not match:
        return None

    name, bindings = match

    jaxes, jhats = get_joypad_axes(device)
    jbuttons = get_joypad_buttons(device)

    axes, buttons = map_gamepad(bindings, jaxes, jbuttons, jhats)

    # Find relevant events for wake-ups
    wake_buttons = [btn for btn, hhd_btn in buttons.items() if hhd_btn in WAKE_KEYS]
    overlay_axes = [
        ax
        for ax, data in axes.items()
        if data.get("code", None) in OVERLAY_AXES
        or data.get("code", data.get("up_code", None)) in OVERLAY_KEYS
    ]
    overlay_buttons = [
        btn for btn, hhd_btn in buttons.items() if hhd_btn in OVERLAY_KEYS
    ]

    return {
        "name": name,
        "guid": guid,
        "axes": axes,
        "buttons": buttons,
        "wake_buttons": wake_buttons,
        "overlay_axes": overlay_axes,
        "overlay_buttons": overlay_buttons,
    }


def test_match_gamepad(mappings):
    jaxes = [i for i in range(20)]
    jbuttons = [i for i in range(500)]
    jhats = [(2 * i, 2 * i + 1) for i in range(10)]

    for mapping in mappings.values():
        name, bindings = mapping
        axes, buttons = map_gamepad(bindings, jaxes, jbuttons, jhats)
        print(f"Checked: {name}")
        print(buttons, axes)
